'''' ---------------------------------------------------------- '''
''' MAIN PROGRAM                                                '''
''' ----------------------------------------------------------- '''
import sys
import os
import scipy
import numpy as np
import pandas as pd
try:
    from harmony import *
    from siman import *
    from threshold import *
    from misc import *
    from ordered_logit_mixed import OrderedMixedLogit
except:
    from .ordered_logit_mixed import OrderedMixedLogit
    from .harmony import *
    from .siman import *
    from .threshold import *
    from .misc import *



''' ----------------------------------------------------------- '''
''' FITTING PARAMETERS TO ORDERED LOGIT                         '''
''' ----------------------------------------------------------- '''
# Assumption: category = {0, 1, ..., J-1}
def ORDLOGMIX(X, y, ncat, normalize=True, start=None, fit_intercept=False):
# {
    mod = OrderedMixedLogit(X=X, y=y, J=ncat, distr='logit', start=start, normalize=normalize, fit_intercept=fit_intercept, varnames = ['vol', 'price', 'carat'])
    mod.fit(start)
    mod.report()
# }


def EXAMPLE_DIAMOND():

    df = pd.read_csv('ord_log_data/diamonds.csv')

    color = ['D', 'E', 'F', 'G', 'H', 'I', 'J']
    df['color'] = pd.Categorical(df['color'], categories=color, ordered=True)
    df['color'] = df['color'].cat.codes

    clarity = ['I1', 'SI1', 'SI2', 'VS1', 'VS2', 'VVS1', 'VVS2']
    df['clarity'] = pd.Categorical(df['clarity'], categories=clarity, ordered=True)
    df['clarity'] = df['clarity'].cat.codes

    df['vol'] = np.array(df['x'] * df['y'] * df['z'])

    cut = ['Fair', 'Good', 'Ideal', 'Premium', 'Very Good']
    df['cut'] = pd.Categorical(df['cut'], categories=cut, ordered=True)
    df['cut'] = df['cut'].cat.codes # Values in {0,1,2,3,4}

    #df.to_csv("diamond_converted.csv", index=False)  # Log revised data to csv file
    varnames = ['vol', 'price', 'carat']
    X = df[['vol', 'price', 'carat']]  # Independent variables
    #X = df[['carat', 'color', 'clarity', 'depth', 'table', 'price', 'vol']]  # Other Independent variables
    y = df['cut']  # Dependent variable
    ncat = 5
    ORDLOGMIX(X, y, ncat, start=None, normalize=True, fit_intercept=False)



if __name__ == '__main__':
    np.random.seed(1)
    EXAMPLE_DIAMOND()





# -*- coding: utf-8 -*-
import pytest

from searchlogit import device

_gpu_available = False
try:
    import cupy
    _gpu_available = True
except ImportError:
    pass


def test_disable_gpu_acceleration():
    """
    Ensure searchlogit properly disables GPU and uses numpy
    """
    device.disable_gpu_acceleration()
    assert not device.using_gpu
    assert device.np.__name__ == 'numpy'


def test_enable_gpu_acceleration():
    """
    Test that when CuPy is not available, it is not possible to enable
    GPU acceleration. This test assumes that CuPy is not installed
    """
    if not _gpu_available:
        with pytest.raises(Exception):
            device.enable_gpu_acceleration()
    else:
        pass


def test_get_device_count():
    """
    Test that when CuPy is not available, 0 GPU devices are returned.
    This test assumes that CuPy is not installed
    """
    count = device.get_device_count()
    if not _gpu_available:
        assert count == 0
    else:
        assert count > 0



"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
IMPLEMENTATION: MULTINOMIAL AND CONDITIONAL LOGIT MODEL 
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

"""
BACKGROUND - MULTINOMIAL LOGIT

The Multinomial Logit Model (MNL) is a statistical model used to analyze categorical
outcome variables with more than two categories. It is a type of regression model 
commonly used in fields such as economics, marketing, transportation planning, 
and social sciences to understand and predict choices among multiple alternatives.

Here's a breakdown of the Multinomial Logit Model:

Dependent Variable: The dependent variable Y represents the categorical outcome with J
categories or alternatives. For example, in a transportation mode choice scenario, the
categories could be car, bus, train, etc. Each individual or observation selects one 
of the available alternatives.

Independent Variables:  In the multinomial logit model, X refers to the data matrix 
of independent variables or predictors. These variables provide information that
influences the choice among the alternatives. These independent variables can be 
categorical or continuous and may represent attributes of the alternatives or 
characteristics of the decision-makers.

Each row of X corresponds to an observation or decision-maker in the dataset. 
For example, if modeling transportation mode choice, each row would represent 
a different traveler.

Each column of X represents a different independent variable or predictor.
These variables can include attributes of the alternatives being considered, 
characteristics of the decision-makers, interactions between variables, 
and other relevant factors.

Utility Function: The choice probability for each alternative is modeled using a 
utility function. The utility U<i,j> of alternative j for individual i is assumed 
to be a linear function of the independent variables:

U<i,j> = B<0,j> + B<1,j> * X<i,1> + B<2,j> * X<i,2> + ... + B<k,j> * X<i,k>

Where:
B<.,j> are the parameters (coefficients) associated with alternative j where . => 1..K
X<i,.> are the independent variables for individual i where . => 1..K
K is the number of independent variables

Choice Probability: The choice probability P<i,j> of individual i selecting alternative j 
is assumed as follows:

P<i,j> = e^U<i,j> /  sum(l=1..J, e^U<i,l>

The softmax function ensures that the choice probabilities sum up to 1 across all
alternatives for each individual.

Estimation: The parameters (coefficients) of the model (B<0,j>, B<1,j>, ... B<k,j>)
are estimated using maximum likelihood estimation (MLE) or other estimation techniques. 
The goal is to find the parameter values that maximize the likelihood of observing 
the actual choices made by individuals in the dataset.

The Multinomial Logit Model assumes independence of irrelevant alternatives (IIA), which means 
that the odds ratio between any two alternatives remains constant regardless of the presence 
of other alternatives. This assumption may not always hold in practice, and alternative 
models such as the Nested Logit Model and the Mixed Logit Model relax this assumption 
to allow for more flexible choice behavior.
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

''' ----------------------------------------------------------- '''
'''  MAIN PARAMETERS:                                           '''
''' ----------------------------------------------------------- '''
# N : Number of choices
# J : Number of alternatives
# K : Number of variables (Kf: fixed (non-trans), Kr: random, Kftrans: fixed trans, Krtrans: random trans)

''' ---------------------------------------------------------- '''
''' UNUSED LIBRARIES                                                  '''
''' ---------------------------------------------------------- '''
#import logging

''' ---------------------------------------------------------- '''
''' LIBRARIES                                                  '''
''' ---------------------------------------------------------- '''
import numpy as np
from scipy.optimize import minimize
from typing import Callable, Tuple
from scipy.optimize import OptimizeResult
import pandas as pd
try:
    from ._choice_model import DiscreteChoiceModel
    from .boxcox_functions import truncate
    from .boxcox_functions import boxcox_param_deriv, boxcox_transformation, truncate_lower
except ImportError:
    from _choice_model import DiscreteChoiceModel
    from boxcox_functions import truncate
    from boxcox_functions import boxcox_param_deriv, boxcox_transformation, truncate_lower


''' ---------------------------------------------------------- '''
''' CONSTANTS - BOUNDS ON NUMERICAL VALUES                     '''
''' ---------------------------------------------------------- '''
max_exp_val = 700
max_val, min_val = 1e+30, 1e-30

''' ---------------------------------------------------------- '''
''' ERROR CHECKING AND LOGGING                                 '''
''' ---------------------------------------------------------- '''
#logger = logging.getLogger(__name__)

''' --------------------------------------------------------------------- '''
''' CLASS FOR MULTINOMIAL AND CONDITIONAL LOGIT MODELS                    '''
''' --------------------------------------------------------------------- '''
class MultinomialLogit(DiscreteChoiceModel):
# {
    """ Docstring """

    # ===================
    # CLASS PARAMETERS
    # ===================
    """
    X:                  Input data for explanatory variables / long format / array-like / shape (n_samples, n_variables)
    y:                  Actual choice made / array-like / shape (n_samples,)
    varnames:           Names of explanatory variables / list / shape (n_variables,)
    alts:               List of alternative names or indexes / long format / array-like / shape (n_samples,)
    isvars:             Names of individual-specific variables in varnames / list
    transvars:          Names of variables to apply transformation on / list / default=None
    transformation:     Transformation to apply to transvars / string / default="boxcox"
    ids:                Identifiers for choice situations / long format / array-like / shape (n_samples,)
    weights:            Weights for the choice situations / long format / array-like / shape (n_variables,) 
                        / default=None
    avail:              Availability indicator of alternatives for the choices (1 => available, 0 otherwise)
                        / array-like / shape (n_samples,)
    base_alt:           Base alternative / int, float or str / default=None
    init_coeff:         Initial coefficients for estimation/ numpy array / shape (n_variables,) / default=None
    bool fit_intercept: Boolean indicator to include an intercept in the model / default=False
    int maxiter:        Maximum number of iterations / default=2000
    float ftol:         Tolerance for scipy.optimize.minimize termination / default=1e-5
    float gtol:         Tolerance for scipy.optimize.minimize(method="bfgs") termination - gradient norm / default=1e-5
    bool return_grad:   Flag to calculate the gradient in _loglik_and_gradient / default=True
    bool return_hess:   Flag to calculate the hessian in _loglik_and_gradient / default=True
    method:             Optimisation method for scipy.optimize.minimize / string / default="bfgs"
    bool scipy_optimisation : Flag to apply optimiser / default=False / When false use own bfgs method.
    
    Assumption: "varnames must match the number and order of columns in X
    """

    # ===================
    # CLASS FUNCTIONS
    # ===================
    """
    1. void __init__(self);
    2. void setup(self, X, y, ...);
    3. void fit(self);
    4. p <-- compute_probabilities(self, betas, X, avail);
    5. Hinv <-- compute_hessian_inverse(self, grad);
    6. result <-- bfgs_optimization(self, betas, X, y, weights, avail, maxiter); 
    7. result <-- scipy_bfgs_optimization(self, betas, X, y, weights, avail, maxiter, ftol, gtol, jac);
    8. loglik <-- get_loglik_null(self); 
    9. loglik <-- validation_loglik(self, validation_X, validation_Y, avail=None, weights=None, betas=None);
    10. (loglik, grad) <-- get_loglik_and_gradient(self, betas, X, y, weights, avail);
    """

    ''' --------------------------------- '''
    ''' Function. Constructor             '''
    ''' --------------------------------- '''
    def __init__(self): # {
        super(MultinomialLogit, self).__init__()  # Base class initialisations
        self.descr = "MNL"
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def compute_hessian_inverse_dummy(self, grad):
        return None

    def compute_hessian_inverse(self, grad):
    # {
        H = np.dot(grad.T, grad)  # Compute dot product

        # Remove zero values and handle NaNs:
        H[np.logical_or(H == 0, np.isnan(H))] = min_val

        # Scale elements if too big or small:
        H = truncate(H, -max_val, max_val)

        Hinv = np.linalg.pinv(H)

        ''' QUERY. IS THIS NECESSARY? WHY NOT CHOOSE ONE APPROACH? '''
        '''try:
            Hinv = np.linalg.inv(H)  # O(n^3) operation
        except Exception:  # use pseudo inverse if normal inverse fails
            Hinv = np.linalg.pinv(H)  # O(n^3) operation'''

        return Hinv
    # }

    ''' -------------------------------------------------------------------- '''
    ''' Function.                                                            '''
    ''' -------------------------------------------------------------------- '''
    def setup(self, X, y, varnames=None, alts=None, isvars=None, transvars=None,
        transformation="boxcox", ids=None, weights=None, avail=None,
        base_alt=None, fit_intercept=False, init_coeff=None, maxiter=2000,
        ftol=1e-5, gtol=1e-5, return_grad=True, return_hess=True,
        method="bfgs", scipy_optimisation=True):
    # {

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # RECAST AS NUMPY NDARRAY
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        X, y, varnames, alts, isvars, transvars, ids, weights, panels, avail = \
            self.set_asarray(X, y, varnames, alts, isvars, transvars, ids, weights, None, avail)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # CHECK FOR MISTAKES IN DATA
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.validate_inputs(X, y, alts, varnames)
        # REMOVED: , isvars, ids, weights, None, base_alt, fit_intercept, maxiter)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # RECORD PARAMETERS AS MEMBER VARIABLES
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.X, self.y = X, y

        if self.X.shape[1] > 0 and self.X[0].dtype == np.object_:
            Xis_numeric = pd.to_numeric(self.X.flatten(), errors='coerce').reshape(self.X.shape)
            Xis_numeric = np.nan_to_num(Xis_numeric, nan=0)
            self.X = Xis_numeric

        self.pre_process(alts, varnames, isvars, transvars, base_alt, fit_intercept, transformation, maxiter, panels)
        self.weights, self.avail = weights, avail
        self.init_coeff = init_coeff
        self.ftol, self.gtol  = ftol, gtol
        self.return_grad, self.return_hess = return_grad, return_hess
        self.method = method
        self.scipy_optimisation = scipy_optimisation

        HessianFunction = Callable[[np.ndarray], np.ndarray]
        self.function_hessian: HessianFunction = self.compute_hessian_inverse_dummy \
            if return_hess else self.compute_hessian_inverse

        Args = Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, int, float, float, bool]
        OptimizeFunction = Callable[[Args], OptimizeResult]
        self.optimizer: OptimizeFunction = self.scipy_bfgs_optimization \
            if scipy_optimisation else self.bfgs_optimization

        self.panels = panels
        self.jac = True if self.return_grad else False
        self.fxidx, self.fxtransidx = [], []  # Initialise empty arrays
        # FIX ME
        for var in self.asvars: # {
            is_transvar = var in self.transvars     # True or False state
            self.fxidx.append(not is_transvar)
            self.fxtransidx.append(is_transvar)
        # }

        if self.transformation == "boxcox": # {
            self.trans_func = boxcox_transformation
            self.transform_deriv = boxcox_param_deriv
        # }

        self.fixedtransvars = self.transvars
        self.X, self.Xnames = self.setup_design_matrix(self.X)

        #expected_length = self.Kf + self.Kftrans * 2 #remove 
        expected_length_betas = self.Kf + self.Kftrans # Add 
        expected_length_lambdas = self.Kftrans # Add 
 
        expected_length = expected_length_betas + expected_length_lambdas #Add 
        if self.init_coeff is None:
            #self.betas = np.repeat(0.1, expected_length)# Remove 
            self.betas = np.append((np.repeat(0.1, expected_length_betas)), np.repeat(1, expected_length_lambdas))# Add
        else: #{
            self.betas = self.init_coeff
            if len(self.init_coeff) != expected_length \
                    and not (hasattr(self, 'is_latent_class') and self.is_latent_class):
                raise ValueError(f"The size of init_coeff must be: {self.X.shape[1]}")
        # }

        if self.weights is not None:
            self.weights = self.weights.reshape(self.X.shape[0], self.X.shape[1]) # Match the size of X

        if self.avail is not None:
            self.avail = self.avail.reshape(self.X.shape[0], self.X.shape[1]) # Match the size of X

        # The code self.y.reshape(-1, self.J) is reshaping the array self.y into a two-dimensional
        # array with a specified number of columns (self.J) while inferring the appropriate number
        # of rows based on the size of the original array.
        self.y = self.y.reshape(-1, self.J)

        self.obs_prob = np.mean(self.y, axis=0)  # Compute: obs_prob[j] = average(y[:,j])
    # }

    ''' -------------------------------------------------------------------- '''
    ''' Function. Fit multinomial and/or conditional logit models            '''
    ''' -------------------------------------------------------------------- '''
    def fit(self, **kwargs):
    # {
        #this is grabbing the wrong X, I am going to try and supply this as an argument, it will only do this if kwargs is degined
        args = (kwargs.get('betas',self.betas), kwargs.get('X',self.X), self.y, self.weights, self.avail,
                self.maxiter, self.ftol, self.gtol, self.jac)
        result = self.optimizer(*args)  # Unpack the tuple and apply the optimizer

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Save predicted and observed probabilities to display in summary
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if 'is_latent_class' not in result.keys(): # i.e., result['is_latent_class'] is not defined
        # {
            p = self.compute_probabilities(result['x'], self.X, self.avail)
            self.ind_pred_prob = p
            self.choice_pred_prob = p
            self.pred_prob = np.mean(p, axis=0) # Compute: pred_prob[j] = average(p[:,j])
        # }

        sample_size = self.X.shape[0]   # Code shortcut for next line
        self.post_process(result, self.Xnames, sample_size)
    # }

    ''' ------------------------------------------------------------ '''
    ''' Function. Compute choice probabilities for each alternative  '''
    ''' ------------------------------------------------------------ '''
    def compute_probabilities(self, betas, X, avail):
    # {
        Xf = X[:, :, self.fxidx]  # Extract a particular matrix
        XB = np.zeros(Xf.shape[0])  # Initialize XB with zeros and the correct shape
        if self.Kf > 0:
            B = betas[:self.Kf]  # Extract first Kf elements of betas
            XB = Xf.dot(B)  # Perform matrix multiplication

        # NOTE: self.fxtransidx is a boolean array
        # and code "sum(self.fxtransidx)" is a proxy for "list(self.fxtransidx).count(True)"

        if any(self.fxtransidx): # {
            KTot = self.Kf + self.Kftrans # Compute the cardinality - scalar
            B_transvars = betas[self.Kf:KTot] # Extract elements [Kf:KTot]
            lambdas = betas[KTot:] # Extract elements [KTot+1:]
            X_trans = X[:, :, self.fxtransidx] # Extract a particular matrix
            XB += self.trans_func(X_trans, lambdas).dot(B_transvars) # Apply transformation and multiply by B_transvars
        # }

        XB = truncate(XB, -max_exp_val, max_exp_val) # Truncate if bound exceeded
        XB = XB.reshape(self.N, self.J)  # Reshape as a matrix - WHY?ALREADY A MATRIX?
        eXB = np.exp(XB)  # Evaluate exponential values!

        if avail is not None:
            eXB *= avail   # i.e., Use in-place multiplication operator. Equivalent to: eXB = eXB * avail

        div = np.sum(eXB, axis=1, keepdims=True)    # Compute the sum along
        p = eXB / div if np.all(div) else np.zeros_like(eXB) # Scale matrix values. Avoid division by zero

        return p  # matrix with dimension ? x ?
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Compute log-likelihood, gradient, and hessian    '''
    ''' ---------------------------------------------------------- '''
    def get_loglik_and_gradient(self, betas, X, y, weights, avail):
    # {
        self.total_fun_eval += 1
        p = self.compute_probabilities(betas, X, avail)

        # ~~~~~~~~~~~~~~~~~~~~~~~
        # Compute Log likelihood
        # ~~~~~~~~~~~~~~~~~~~~~~~
        lik = np.sum(y * p, axis=1)     # Sum the elements across axis = 1
        lik = truncate_lower(lik, min_val)    # Remove zero values, i.e., lik[lik == 0] = min_val
        loglik = np.log(lik)            # Log each element
        if weights is not None:
            loglik = loglik * weights[:, 0]  # doesn't matter which column
        loglik = np.sum(loglik)         # Sum up the elements

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Individual contribution to the gradient
        # and position of trans vars -  B_trans, lambdas, X_trans
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        Kmax = self.Kf + self.Kftrans
        ymp = y - p

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Compute grad
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        grad = np.einsum('nj,njk -> nk', ymp, X[:, :, self.fxidx]) if self.Kf > 0 else np.zeros((ymp.shape[0], 0))

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Compute individual contribution of trans to the gradient
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if self.Kftrans > 0: # {
            transpos = [self.varnames.tolist().index(i) for i in self.transvars]
            X_trans = X[:, :, transpos]
            X_trans = X_trans.reshape(self.N, len(self.alts), len(transpos))
            lambdas = betas[Kmax:]  # Extract elements [Kmax+1:]
            Xtrans_lmda = self.trans_func(X_trans, lambdas)
            gtrans = np.einsum('nj,njk -> nk', ymp, Xtrans_lmda)
            der_Xtrans_lmda = self.transform_deriv(X_trans, lambdas)
            B_trans = betas[self.Kf:Kmax]  # Extract elements [Kf:Kmax]
            der_XBtrans = np.einsum('njk,k -> njk', der_Xtrans_lmda, B_trans)
            gtrans_lmda = np.einsum('nj,njk -> nk', ymp, der_XBtrans)
            grad = np.concatenate((grad, gtrans, gtrans_lmda), axis=1) if grad.size \
                else np.concatenate((gtrans, gtrans_lmda), axis=1)  # (N, K)
        # }

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Assuming grad.shape = (m, n) and weights.shape = (m, k)
        if weights is not None:
            # Compute grad = np.transpose(np.transpose(grad) * weights[:, 0])
            # The code below takes the first column of the weights array and reshapes it into a column vector.
            # The reshape converts weights_first_column into a 2d array with a single column.
            weights_first_column = weights[:, 0] # Select first column for element-wise multiplication
            grad *= weights_first_column.reshape(-1, 1) # Element-wise multiplication using broadcasting
            # The "*=" operation multiplies each column of grad with weights_first_column
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        self.Hinv = self.function_hessian(grad)  # Conditionally compute Hinv based upon return_hess flag
        grad = np.sum(grad, axis=0)     # Compute grad_[i] = sum(j, grad[i][j])
        self.gtol_res = np.linalg.norm(grad, ord=np.inf)  # Compute the norm of "grad"

        penalty = self.regularize_loglik(betas)
        loglik = loglik - penalty
        
        result = (-loglik, -grad, self.Hinv) if self.return_grad and self.return_hess \
            else (-loglik, -grad) if self.return_grad else (-loglik,)
        return result
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Compute the log-likelihood on the validation set '''
    ''' using the betas fitted using the training set              '''
    ''' ---------------------------------------------------------- '''
    def get_validation_loglik(self, validation_X, validation_Y, avail=None, weights=None, betas=None):
    # {
        validation_X, Xnames = self.setup_design_matrix(validation_X)
        validation_Y = validation_Y.reshape(self.N, self.J)
        betas = betas if betas is not None else self.coeff_est
        res = self.get_loglik_and_gradient(betas, validation_X, validation_Y, avail=avail, weights=weights)
        loglik = res[0]
        return loglik
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def scipy_bfgs_optimization(self, betas, X, y, weights, avail, maxiter, ftol, gtol, jac):
    # {
        args = (X, y, weights, avail)
        options = {'gtol': gtol, 'maxiter': maxiter, 'disp': False}
        result = minimize(self.get_loglik_and_gradient, betas,
                args=args, jac=jac, method=self.method, tol=ftol, options=options)
        return result
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. This code is an implementation of the            '''
    ''' Broyden–Fletcher–Goldfarb–Shanno (BFGS) algorithm update   '''
    ''' rule for the inverse Hessian matrix (Hinv).                 '''
    ''' ---------------------------------------------------------- '''
    def bfgs_optimization(self, betas, X, y, weights, avail, maxiter, ftol, gtol, jac):
    # {
        res, g, Hinv = self.get_loglik_and_gradient(betas, X, y, weights, avail)
        current_iteration = 0
        resnew, gnew, s = 0, 0, 0

        while True:
        # {
            old_g = g
            Hinv_g = -Hinv.dot(g)

            #  This code is a loop that iteratively adjusts the step size and updates
            #  the betas variable until a certain condition is met.
            step = 2
            max_iterations = 1000  # Define a maximum number of iterations to prevent infinite loops
            for _ in range(max_iterations): # {
                step /= 2       # i.e. step = step / 2
                s = step * Hinv_g
                betas += s
                resnew, gnew, _ = self.get_loglik_and_gradient(betas, X, y, weights, avail)
                if resnew <= res or step < 1e-10: break
            else:
                print("Maximum iterations reached without convergence.")
            # }

            old_res = res
            res = resnew
            g = gnew

            # Compute intermediate terms
            delta_g = g - old_g
            s_dot_delta_g = s.dot(delta_g)
            delta_g_Hinv = delta_g[None, :].dot(Hinv)
            s_outer_s = np.outer(s, s)

            # Compute the first term
            first_term_numerator = s_dot_delta_g + delta_g_Hinv.dot(delta_g)
            first_term = (first_term_numerator * s_outer_s) / (s_dot_delta_g ** 2)

            # Compute the second term
            Hinv_dot_delta_g = Hinv.dot(delta_g)
            second_term = (np.outer(Hinv_dot_delta_g, s) + np.outer(s, delta_g_Hinv)) / s_dot_delta_g

            # Update Hinv
            Hinv += first_term - second_term

            # Evaluate termination conditions
            current_iteration = current_iteration + 1  # Increment counter
            converged = (abs(res - old_res) < 0.00001)
            limit_reached = (current_iteration > maxiter)
            if (converged or limit_reached): break

        # }

        result  = {'success': converged, 'x': betas, 'fun': res, 'hess_inv': Hinv, 'nit': current_iteration}
        return result
    # }
# }


"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
IMPLEMENTATION: ORDERED LOGIT
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
'''
THEORY: Ordered Logit models the cumulative probabilities of being in a 
particular category or higher (or lower) based on thresholds. It assumes
 an underlying continuous latent variable that determines the category.

Ordered Logit assumes the Proportional Odds assumption, which means the 
effect of the independent variables is assumed to be the same for each 
threshold (i.e., across all categories).

NOTATION: 
N - number of observations (a.k.a., samples)
J - number of categories (a.k.a., alternatives)
K - number of features (a.k.a., predictor variables)
Y - observed category (a.k.a., ordinal variable); |Y| = N
Y* - unobserved dependent (latent) variable (continuous); |Y*| = Nx1
j - category index; j = 1,...,J
X - Explanatory variables; |X| = N x K
thr - thresholds between categories; |thr| = J-1
err - error term; follows logistic distribution; |err| = N
params - vector of parameters to be estimated; |params| = (K + 1)  + (J - 1)
param = [beta[0], beta[1],...,beta[K], thr[1], thr[2],..., thr[J-1]]

Assumption: Y is a function of Y*
Y[i] = 1 if Y*[i] <= thr[1]
Y[i] = 2 if thr[1] < Y*[i] <= thr[2]
Y[i] = 3 if thr[2] < Y*[i] <= thr[3]
...
Y[i] = J if Y*[i] > thr[J-1] 

Note: Y* is not measured
Assumption: Y* = X.beta + err
Y*[i] = sum(j=1,..,J: X[i,j].beta[j] + err[i]
 
Let Z[i] = sum(j=1,..,J: X[i,j].beta[j] = E(Y*[i])

P(Y=j|X) = P(thr[j-1] < Y* <= thr[j])
         = cdf(thr[j] - X.beta) - cdf(thr[j-1] - X.beta)

?:
cdf(p) = ln(p/(1-p))  
cdf(p) = 1 / (1 + e^-p) 
cdf(p) = e^p / (1 + e^p)

GOAL: Identify optimal thr and betas 

'''

''' ---------------------------------------------------------- '''
''' LIBRARIES                                                  '''
''' ---------------------------------------------------------- '''
import numpy as np
from _choice_model import DiscreteChoiceModel

from scipy import stats
from scipy.optimize import minimize

minval = 1E-30

# Extract first n elements of the array
# Note: This function provides a "in-place" reference
def get_first_elements(arr: np.ndarray, n)->np.ndarray:
    return arr[:n]

# Extract last n elements of the array
# Note: This function provides an "in-place" reference
def get_last_elements(arr: np.ndarray, n)->np.ndarray:
    return arr[-n:]

# Replace last n elements of the array
def set_last_elements(arr: np.ndarray, n, sub_arr: np.ndarray):
    arr[-n:] = sub_arr

def set_first_elements(arr: np.ndarray, n, sub_array: np.ndarray):
    arr[:n] = sub_array

''' ---------------------------------------------------------- '''
''' Function. Perform Cholesky decomposition on the Hessian 
Assumes H is positive-definite.                                '''
''' ---------------------------------------------------------- '''
def cholesky_decomposition(H):
# {
    try:
        L = np.linalg.cholesky(H)
        return L
    except np.linalg.LinAlgError:
        print("Matrix is not positive-definite")
        return None
# }

def compute_inverse_cholesky(H):
# {
    L = cholesky_decomposition(H)
    if L is not None:
        # L is the lower triangular matrix, so we need to solve for L^-1
        L_inv = np.linalg.inv(L)  # Inverse of lower triangular matrix L

        # H^-1 = (L^-1)^T @ L^-1
        H_inv = np.dot(L_inv.T, L_inv)  # (L^-1)^T * L^-1
        return H_inv
    else:
        return None
# }

''' ---------------------------------------------------------- '''
''' CLASS FOR ESTIMATION OF ORDERED LOGIT                      '''
''' ASSUMPTION: ALL DATA SHOULD BE NORMALISED                  '''
''' ---------------------------------------------------------- '''
class OrderedLogit():
# {
    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def __init__(self, **kwargs):
    # {
        self.descr = "ORL"
        self.setup(**kwargs)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def normalize_data(self):
    # {
        self.normalize = True
        self.min_data = self.X.min(axis=0)
        self.max_data = self.X.max(axis=0)
        self.range_of_data = self.max_data - self.min_data
        self.X = (self.X - self.min_data) / self.range_of_data
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def standardize_data(self):
    # {
        self.X_mean = self.X.mean(axis=0)
        self.X_std_dev = self.X.std(axis=0)
        self.X = (self.X - self.X_mean) / self.X_std_dev
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. names = (1/2, 2/3, ... J-1/J)                    '''
    ''' ---------------------------------------------------------- '''
    def define_labels(self):
    # {
        self.labels = ["constant"] if self.fit_intercept else ["na"]
        self.labels += [self.varnames[i] for i in range(0, self.K)]
        self.labels += ['threshold: ' + str(i) + '/' + str(i+1) for i in range(1, self.J)]
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Set up the model                                 '''
    ''' ---------------------------------------------------------- '''
    def setup(self, **kwargs):
    # {
        # Assumption - X & y must be dataframes
        self.X = kwargs.get('X')  # The full dataframe
        self.varnames = self.X.columns.tolist()
        self.X = np.asarray(self.X)  # The explanatory variables only
        self.y = kwargs.get('y')  # The full dataframe
        self.y = np.asarray(self.y)  # The observed ordinal values only

        self.N = self.X.shape[0]    # Number of observations (samples)
        self.K = self.X.shape[1]    # Number of predictor variables
        self.J = kwargs.get('J')    # Number of ordinal categories => categories = {0, 1, ..., J-1}
        self.nparams = self.K + self.J  # i.e., intercept + self.K + self.J - 1

        self.params = kwargs.get('start')
        if self.params is None:
            self.params = np.zeros(self.nparams, dtype=float)

        # Outputs
        self.y_latent = np.zeros(self.N, dtype=float)
        self.stderr = np.zeros(self.nparams)
        self.signif_lb = np.zeros(self.nparams)
        self.signif_ub = np.zeros(self.nparams)
        self.pvalues = np.zeros(self.nparams)
        self.zvalues = np.zeros(self.nparams)

        # Undefined:
        self.loglike = None
        self.aic = None
        self.bic = None
        self.method = None

        self.normalize = kwargs.get('normalize')
        if self.normalize: self.normalize_data()

        distr = kwargs.get('distr')
        if distr == 'probit':
            self.distr = stats.norm
        elif distr == 'logit':
            self.distr = stats.logistic
        else:
            self.distr = distr

        self.fit_intercept = kwargs.get('fit_intercept')   # Add intercept
        self.define_labels()
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Extract thresholds -> last self.J - 1 elements   '''
    ''' ---------------------------------------------------------- '''
    def get_thresholds(self, params: np.ndarray)->np.ndarray:
    # {
        delta = get_last_elements(params, self.J - 1)
        thresholds = np.cumsum(delta)
        return thresholds
    # }

    def set_thresholds(self, values: np.ndarray):
        set_last_elements(self.params, self.J - 1, values)

    ''' ---------------------------------------------------------- '''
    ''' Function. Extract betas from the params array              '''
    ''' ---------------------------------------------------------- '''
    def get_beta(self, params: np.ndarray)->np.ndarray:
        return get_first_elements(params, 1 + self.K) # Return params[0],...,params[self.K]

    def set_beta(self, beta:np.ndarray):
        set_first_elements(self.params, 1 + self.K, beta)

    ''' ---------------------------------------------------------- '''
    ''' Function. Linear prediction of latent variable             '''
    ''' i.e., beta[0] + compute X.beta + offset '''
    ''' |X| = N x K and |beta| = K                                 '''
    ''' ---------------------------------------------------------- '''
    def compute_latent(self, X: np.ndarray, beta: np.ndarray)-> np.ndarray:
    # {
        y_latent = X.dot(beta[1:])  # Compute dot product, i.e., X.beta
        if(self.fit_intercept):
            y_latent += beta[0] # Add beta[0] to each y_latent value, i.e. y_latent[n] += beta_0 for n=1,...,N
        return y_latent
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function: Find which category is associated with each value'''
    ''' category[n] = j if values[n] <= thresholds[j+1] and        '''
    ''' values[n] > thresholds[j] for n in [0..N-1]                '''
    ''' ---------------------------------------------------------- '''
    def find_category(self, values: np.ndarray, thresholds: np.ndarray)->np.ndarray:
    # {
        category = np.digitize(values, thresholds)
        return category - 1
    # Note: Subtract - 1 to convert indexing starting from 1 to indexing starting from 0
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Compute ordinal outcome. Y= F(Y*, thresholds)    '''
    ''' ---------------------------------------------------------- '''
    def predict_category(self, y_latent: np.ndarray, thresholds: np.ndarray)->np.ndarray:
         return self.find_category(y_latent, thresholds)

    ''' ---------------------------------------------------------- '''
    ''' Function. Predicted probabilities for each categorical level
    i.e., Compute P(Y=j|X) = P(thr[j-1] < Y* <= thr[j])  for all j   
    Note: Given Y* = X.beta, then: P(Y=j|X) = cdf(thr[j] - X.beta) - cdf(thr[j-1] - X.beta)    
    Note: Define cut = [-inf, thr[1], thr[2], ... thr[J-1], inf] 
    Hence, |cut| = self.J + 1           
    Compute: prob[n][j] = self.prob_interval(cut[j] - y_latent[n], cut[j+1] - y_latent[n])
    for n, j in np.ndindex(self.N, self.J)
    '''
    ''' ---------------------------------------------------------- '''
    def compute_probability(self, X: np.ndarray, params: np.ndarray)-> np.ndarray:
    # {
        thresholds = self.get_thresholds(params)
        beta = self.get_beta(params)
        self.y_latent = self.compute_latent(X, beta)   # y_latent = X.beta; |y_latent| = N x 1

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Vectorized approach, using broadcasting
        # Note: [:-1] => Ignore last element
        # Note: [1:]  => Ignore first element
        # Note: |cut[:-1]| = |cut[1:]| = self.J and |low| = |high| = self.N x self.J
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        cut = np.concatenate(([-np.inf], thresholds, [np.inf]))
        y_latent = self.y_latent[:, None]  # Add a second dimension => |y_latent| = (self.N, 1)
        low = cut[:-1] - y_latent  # Compute: cut[j] - y_latent[n] for j in range(0, self.J - 1)
        high = cut[1:] - y_latent  # Compute: cut[j + 1] - y_latent[n] for j in range(1, self.J - 1)
        prob = self.prob_interval(low, high) # Note: |prob| = self.N x self.J
        return prob
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. CDF and PDF calculations                         '''
    ''' ---------------------------------------------------------- '''
    def cdf(self, value: np.ndarray)-> np.ndarray:
        return self.distr.cdf(value)

    def pdf(self, value: np.ndarray)-> np.ndarray:
        return self.distr.pdf(value)

    ''' ---------------------------------------------------------- '''
    ''' Function. Interval probability                             '''
    ''' Probability that a value falls within interval (low, up]   '''
    ''' ---------------------------------------------------------- '''
    def prob_interval(self, low: np.ndarray, high: np.ndarray)-> np.ndarray:
        p = self.cdf(high) - self.cdf(low)
        return p

    ''' ---------------------------------------------------------- '''
    ''' Function. This function calculates how likely it is to observe 
    the actual ordinal category for each individual data point, given
    the model parameters (including coefficients and thresholds).
    The likelihood is a product of probabilities across observations, hence we
    take the log of each probability to form the log-likelihood.
    Note: p[n][j] = P(Y[n] = j)
    
    like(Y[n] |X, beta, thresholds) = P(Y[n] = y_obs[n])
    '''
    def get_loglike_obs(self, params: np.ndarray)->np.ndarray:
    # {
        p = self.compute_probability(self.X, params) # |p| = N x J
        p = np.clip(p, minval, 1.0)  # Elements: Force value to range (0, 1]
        like = np.array([p[n][self.y[n]] for n in range(self.N)])
        return np.log(like)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def get_loglike(self, params: np.ndarray)->float:
        loglike_obs = self.get_loglike_obs(params)  # Log-likelihood for each observation
        loglik = np.sum(loglike_obs)
        return loglik

    ''' ---------------------------------------------------------- '''
    ''' Function. Change the predictors and hence K, X, names      '''
    ''' ---------------------------------------------------------- '''
    def revise_specification(self, K, X, names):
    # {
        self.K = K
        self.X = X
        self.nparams = self.K + self.J
        self.names = names
        self.define_labels()
        self.params = np.zeros(self.nparams)
        self.stderr = np.zeros(self.nparams)
        self.signif_lb = np.zeros(self.nparams)
        self.signif_ub = np.zeros(self.nparams)
        self.pvalues = np.zeros(self.nparams)
        self.zvalues = np.zeros(self.nparams)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    # Objective is to maximize the score
    # minimize = True => Transform maximization to equivalent minimization objective
    def evaluate(self, params: np.ndarray, minimize=True)->float:
    # {
        self.loglike = self.get_loglike(params)
        score = self.loglike
        return -score if minimize else score
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def get_loglike_gradient(self, params: np.ndarray, delta: np.ndarray):
    # {
        score = self.evaluate(params)
        gradient = self.compute_gradient_central(params, delta)
        return (score, gradient)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def compute_gradient(self, params:np.ndarray, delta:np.ndarray)->np.ndarray:
    # {
        gradient = np.zeros_like(params)  # create an array
        f = self.evaluate(params)
        for i in range(len(params)):
        # {
            orig = params[i]
            params[i] = orig + delta[i] # perturb value
            f_plus = self.evaluate(params)
            params[i] = orig # restore value
            gradient[i] = (f_plus - f) / (delta[i])
        # }
        return gradient
    # }

    def compute_gradient_central(self, params:np.ndarray, delta:np.ndarray)->np.ndarray:
    # {
        gradient = np.zeros_like(params) # create an array
        for i in range(len(params)):
        # {
            orig = params[i]
            params[i] = orig + delta[i]
            case_1 = self.evaluate(params)
            params[i] = orig - delta[i]
            case_2 = self.evaluate(params)
            params[i] = orig # restore value
            gradient[i] = (case_1 - case_2) / (2.0 * delta[i])
        # }
        return gradient
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def get_hessian(self, eps=1e-6):
    # {
        N = self.nparams  # Cardinality of hessian matrix
        hessian = np.zeros((N, N))  # Initialise hessian matrix
        delta = [eps] * N
        params = np.copy(self.params)
        df_0 = self.compute_gradient_central(params, delta)
        for i in range(N):  # i.e., for i = 0, 1, ..., N-1
        # {
            params[i] += eps  # Increment by epsilon
            df_1 = self.compute_gradient_central(params, delta)
            hessian[i, :] = (df_1 - df_0) / eps  # Compute the gradient for row i elements
            params[i] -= eps  # Undo the change
        # }
        return hessian

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def get_hessian_2(self, eps=1e-6):
        N = self.nparams  # Cardinality of hessian matrix
        hessian = np.zeros((N, N))  # Initialise hessian matrix

        for i in range(N):
            for j in range(i, N):
                params = np.copy(self.params)  # Create a copy of the parameters
                # Perturb the parameters in all four combinations
                params[i] += eps
                params[j] += eps
                case_1 = self.evaluate(params)
                params[j] -= 2.0 * eps
                case_2 = self.evaluate(params)
                params[i] -= 2.0 * eps
                case_4 = self.evaluate(params)
                params[j] += 2.0 * eps
                case_3 = self.evaluate(params)
                # Compute the second-order mixed partial derivative for hessian[i, j]
                hessian[i, j] = (case_1 - case_2 - case_3 + case_4) / (4.0 * eps ** 2.0)
                hessian[j, i] = hessian[i, j]
        return hessian
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def get_hessian_central(self, eps=1e-6):
    # {
        N = self.nparams  # Cardinality of Hessian matrix
        hessian = np.zeros((N, N))  # Initialize Hessian matrix
        params = np.copy(self.params)
        delta = [eps] * N

        for i in range(N):  # Iterate over parameters to compute second derivatives
            # Perturb parameter i positively and negatively by eps
            params[i] += eps
            df_pos = self.compute_gradient_central(params, delta)

            params[i] -= 2.0 * eps  # Perturb parameter i negatively by 2*eps
            df_neg = self.compute_gradient_central(params, delta)

            # Compute second derivative using central difference
            hessian[i, :] = (df_pos - df_neg) / (2.0 * eps)

            # Reset parameter i to original value
            params[i] += eps

        return hessian
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def get_bic(self, loglike):
        return np.log(self.N) * self.nparams - 2.0 * loglike

    def get_aic(self, loglike):
         return 2.0 * self.nparams - 2.0 * loglike

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    ''' Compute the standard errors - a measure of the variability 
        or uncertainty of a sample statistic. It describes how much the sample statistic 
        is expected to vary from the true population parameter. In other words, the 
        standard error provides an indication of the precision of an estimate.  
        SMALLER => BETTER ESTIMATE    
    '''
    def compute_stderr(self, tol):
    # {
        hessian = self.get_hessian(tol)
        inverse = np.linalg.pinv(hessian) # Conventional approach
        diag = np.diagonal(inverse)
        diag_copy = np.copy(diag)
        diag_copy[diag_copy < minval] = 0

        # DEBUG:
        #for i, value in enumerate(diag_copy):
        #    if value < 0:
        #        diag_copy[i] = 0

        # Standard errors are the square root of the diagonal elements of the variance-covariance matrix
        self.stderr = np.sqrt(diag_copy)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    ''' Compute a 95% confidence interval for each coefficient. 
        Identify the range within which the true coefficient is likely to 
        lie with a given confidence level..
    '''
    def compute_confidence_intervals(self):
        self.signif_lb = self.params - 1.96 * self.stderr # i.e. signif_lb[i] = params[i] - 1.96 * stderr[i]
        self.signif_ub = self.params + 1.96 * self.stderr # i.e.,signif_ub[i] = params[i] + 1.96 * stderr[i]

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    ''' The z-value is used to test the null hypothesis that the coefficient is equal
        to zero (i.e., there is no effect).
        It represents the number of standard deviations a particular data point is away from the mean of a distribution
        BELOW 1.96 => NOT STATISTICALLY SIGNIFICANT 
    '''
    def compute_zvalues(self):
    # {
        for i in range(self.nparams):
        # {
            if self.stderr[i] > minval:
                self.zvalues[i] = self.params[i] / self.stderr[i]
            else:
                self.zvalues[i] = np.nan
        # }
        self.zvalues = np.clip(self.zvalues, -np.inf, np.inf)  # Set limits
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    ''' NULL HYPOTHESIS: NO EFFECT OR NO DIFFERENCE, I.E, COEFFICIENT IS ZERO'''
    ''' < 0.05 => REJECT NULL HYPOTHESIS. IT DOES HAVE A SIGNIFICANT EFFECT '''
    ''' > 0.05 => FAIL TO REJECT NULL HYPOTHESIS. IT IS UNLIKELY TO HAVE A SIGNIFICANT EFFECT'''
    def compute_pvalues(self):
    # {
        if self.nparams < 100:
            self.pvalues = 2.0 * (1.0 - stats.t.cdf(np.abs(self.zvalues), df=self.nparams))
        else:
            self.pvalues = 2.0 * (1.0 - stats.norm.cdf(np.abs(self.zvalues)))
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def post_process(self):
    # {
        self.loglike = self.evaluate(self.params, False)
        self.aic = self.get_aic(self.loglike)
        self.bic = self.get_bic(self.loglike)
        self.compute_stderr(1E-4) # The tolerance can be temperamental. USe 1E-2
        self.compute_zvalues()
        self.compute_pvalues()
        self.compute_confidence_intervals()

        if self.normalize:
            self.unscale_param()
            print("Beta(unscaled) =",self.unscaled_beta)
            print("Threshold(unscaled) =",self.unscaled_threshold)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def unscale_param(self):
    # {
        self.unscaled_beta = self.unscale_beta()
        self.unscaled_threshold  = self.unscale_threshold()
    # }

    def unscale_beta(self):
    # {
        beta = (self.get_beta(self.params)).copy()
        for k in range(0, self.K):
            beta[k+1] = beta[k+1] / self.range_of_data[k]
        return beta
    # }

    def unscale_threshold(self):
    # {
        beta = (self.get_beta(self.params)).copy()
        threshold = self.get_thresholds(self.params).copy()
        offset = sum(beta[k+1] * self.min_data[k] / self.range_of_data[k] for k in range(self.K))
        threshold += offset  # i.e., threshold[j] += sum for j in range(self.J):
        return threshold
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. '''
    ''' ---------------------------------------------------------- '''
    def report(self):
    # {
        print("=" * 110)
        print("Method: ",self.method)
        print("Log-Likelihood: {:.5f}".format(self.loglike))
        print("AIC: {:.5f}".format(self.aic))
        print("BIC: {:.5f}".format(self.bic))
        print("=" * 110)

        # Print out table:
        print("{:>20} {:>12} {:>12} {:>12} {:>12} {:>12} {:>12}"
        .format("Coeff","Estimate","Std.Err.","z-val","p-val","[0.025","0.975]"))
        print("-" * 110)
        cond = "{:>20} {:>12.6f} {:>12.6f} {:>12.6f} {:>12.6f} {:>12.6f} {:>12.6f}"
        lb = 0 if self.fit_intercept else 1

        thr = self.get_thresholds(self.params)
        beta = self.get_beta(self.params)
        params = np.concatenate((beta, thr))

        for i in range(lb, self.nparams):
        # {
            formatted_str = cond.format(self.labels[i], params[i], self.stderr[i],
                self.zvalues[i], self.pvalues[i], self.signif_lb[i], self.signif_ub[i])
            if self.pvalues[i] < 0.05:
                formatted_str += (" (*)")
            print(formatted_str)
        # }
        print("=" * 110)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Fit the model                                    '''
    ''' ---------------------------------------------------------- '''
    def fit(self, method='L-BFGS-B', start=None):
    # {
        options = {'maxiter': 50000}
        tol = 1e-8 # tol = 1e-5
        self.method = method

        if start is None:
            start = [0] * self.params
            value = [1] * (self.J - 2) # These are the deltas
            set_last_elements(start, self.J - 2, value)

        delta = np.ones(self.nparams) * tol
        bounds_beta = [(-np.inf, np.inf)] * (self.K + 2)  # K+1 betas + 1 threshold. [-inf, inf]
        bounds_delta = [(minval, np.inf)] * (self.J - 2)  # These are deltas. [0, inf]
        bounds = np.concatenate((bounds_beta, bounds_delta))
        args = (delta,)  # Make sure this is a tuple by adding a comma
        optimize_result = minimize(fun=self.get_loglike_gradient, x0=start, args=args,
            method='L-BFGS-B', tol=tol, jac=True, options=options, bounds=bounds)

        self.params = optimize_result.x  # Extract results
        self.post_process()
    # }
# }


''' ---------------------------------------------------------- '''
''' METAH                                                      '''
''' ---------------------------------------------------------- '''
# Maximize objective assumed!
def accept_change(current: float, proposed: float, t: float):
    if proposed > current:
        return True
    else:
        delta = proposed - current  # delta > 0 => How much better the solution is
        ln_prob = delta / t
        return np.log(np.random.random()) < ln_prob
# }


# MAY NOT BE VALID FOR NEW THRESHOLDING APPROACH - TO CHECK !!!!!!!!
class SA_ORDLOG_FIT():
# {
    def __init__(self, mod: OrderedLogit, maxiter: int):
    # {
        self.mod = mod
        self.step = 0
        self.tI = 10
        self.tF = 0.001
        self.t = self.tI
        self.rate = np.exp((1.0 / (maxiter - 1)) * np.log(self.tF / self.tI));
        self.maxiter = maxiter
        self.improved_best =  False

        # Starting beta. Set beta[1], beta[2], ... as 1
        self.current = np.zeros_like(mod.params)
        self.current[1: self.mod.K + 1] = 1  # Index: [1], [2], ..., [K]

        # Starting threshold. Set thresholds as [0, delt, 2*delt, ...]
        # for index: [K+1], [K+2], ..., [N-1]
        N = self.mod.nparams
        delta = 0.1
        temp = 0
        for i in range(self.mod.K+1, N):
            self.current[i] = temp
            temp += delta

        self.best = np.copy(self.current)
        self.current_score = self.evaluate(self.current)
        self.best_score = self.current_score
    # }

    def restore_best(self):
        self.current = np.copy(self.best)
        self.current_score = self.best_score

    def perturb_beta(self):
    # {
        cand = np.copy(self.current)
        ub = 1 + self.mod.K
        index = np.random.randint(0, ub)
        delta = np.random.uniform(-2,2)
        cand[index] += delta
        candidate_score = self.evaluate(cand)
        if accept_change(self.current_score, candidate_score, self.t):
            self.current = np.copy(cand)
            self.current_score = candidate_score
            self.update_best()
    # }

    def perturb_threshold(self):
    # {
        cand = np.copy(self.current)
        ub = len(cand)
        lb = ub - self.mod.J + 1
        index = np.random.randint(lb, ub)
        delta = np.random.uniform(-1,1)
        cand[index] += delta

        # Perform correction:
        for i in range(lb, ub-1):
        # {
            if cand[i] > cand[i+1]:
                temp = cand[i]
                cand[i] = cand[i+1]
                cand[i+1] = temp
        # }

        score = self.evaluate(cand)
        if accept_change(self.current_score, score, self.t):
        # {
            self.current = cand
            self.current_score = score
            self.update_best()
        # }
    # }

    def local_search(self, eps=0.1, max_iter=1):
    # {
        step_size = 0.1
        for _ in range(max_iter):
        # {
            for i in range (len(self.current)):
            # {
                param = np.copy(self.current)
                param[i] += step_size
                score = self.evaluate(param)
                if score < self.current_score:
                # {
                    self.current = param
                    self.current_score = score
                    self.update_best()
                # }

                param = np.copy(self.current)
                param[i] -= step_size
                if score < self.current_score:
                # {
                    self.current = param
                    self.current_score = score
                    self.update_best()
                # }
            # }
        # }
    # }

    def evaluate(self, solution):
        return self.mod.evaluate(solution, False)

    def update_best(self):
    # {
        if self.current_score > self.best_score:
            self.best_score = self.current_score
            self.best = self.current
            self.improved_best = True
    # }

    def run(self):
    # {
        no_impr = 0
        for iter in range(self.maxiter):
        # {
            self.perturb_beta()
            self.perturb_threshold()
            if iter % 200 == 0: self.local_search()

            # Optional:
            #if self.improved_best == False:
            #    no_impr += 1

            #if no_impr >= 10:
            #    self.restore_best()
            #    no_impr = 0

            self.t = self.t * self.rate  # Revise temperature
            self.step += 1

            print("step=",self.step,"; t=", self.t,"; current=",self.current_score, "(best=", self.best_score,")")
        # }
        self.restore_best()
        return self.best, self.best_score
    # }
# }


''' ---------------------------------------------------------- '''
''' METAH                                                      '''
''' ---------------------------------------------------------- '''
class SA_ORDLOG():
# {
    def __init__(self, X, y, J, maxiter: int):
    # {
        # Ordered logit parameters
        self.varnames = X.columns.tolist()
        self.X = np.asarray(X)  # The explanatory variables only
        self.y = y
        self.J = J
        self.N = X.shape[0]  # Total number of observations (samples)
        self.K = X.shape[1]  # Total number of predictors
        self.mod = OrderedLogit(X=X, y=y, J=J, distr='logit', start=None, fit_intercept=True)

        # SA specific parameters
        self.archive = {}  # Define a dictionary
        self.tI = 100
        self.tF = 0.001
        self.t = self.tI
        self.rate = np.exp((1.0 / (maxiter - 1)) * np.log(self.tF / self.tI));
        self.maxiter = maxiter
        self.current = np.zeros(self.mod.K, dtype=int)  # Current selection of predictors; 1 => include, 0 => exclude
        self.best = np.copy(self.current)
        self.current_score = self.evaluate(self.current)
        self.best_score = self.current_score
    # }

    def evaluate(self, solution):
    # {
        chosen = [index for index, value in enumerate(solution) if value == 1]
        X = self.X[:, chosen]  # Grab specific columns of self.X
        names = [self.varnames[i] for i in chosen]# Grab specific names
        self.mod.revise_specification(len(chosen), X, names) # Revise the ordered logit model
        self.mod.fit()
        return self.mod.loglike
    # }

    def restore_best(self, eval=True):
        self.current = np.copy(self.best)
        self.current_score = self.best_score
        self.evaluate(self.current)

    def update_best(self):
        if self.current_score > self.best_score:
            self.best_score = self.current_score
            self.best = np.copy(self.current)

    def perturb(self):
    # {
        cand = np.copy(self.current)

        pert = np.random.randint(3) # Choose a perturbation type
        if pert == 0: # FLip
            i = np.random.randint(0, self.K)
            cand[i] = 1 - cand[i] # Flip 0 to 1 or 1 to 0

        if pert == 1: # Add
            options = [i for i in range(self.K) if cand[i] == 0]
            if len(options) > 0:
                i = np.random.randint(0, len(cand))
                cand[i] = 1

        if pert ==2: # Remove
            options = [i for i in range(self.K) if cand[i] == 1]
            if len(options) > 0:
                i = np.random.randint(0, len(cand))
                cand[i] = 0

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        key = tuple(cand)  # Hash the array
        if key not in self.archive:
            candidate_score = self.evaluate(cand)
            self.archive[key] = self.mod.loglike  # Record array
        else:
            print("Solution already seen")
            candidate_score = self.archive[key]

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        if accept_change(self.current_score, candidate_score, self.t):
        # {
            self.current = np.copy(cand)
            self.current_score = candidate_score
            self.update_best()
        # }
    # }

    def run(self):
    # {
        noimpr = 0
        for iter in range(self.maxiter):
        # {
            prev_best = self.best_score
            self.perturb()
            self.t = self.t * self.rate  # Revise temperature
            nb = np.sum(self.current)
            nb_best = np.sum(self.best)
            message = "t={:.5f}: current={:.5f}({}), best={:.5f}({})".format(self.t,self.current_score,nb,self.best_score,nb_best)
            print(message)

            if self.best_score <= prev_best: noimpr += 1
            if noimpr >= 10:
                print("restore best")
                self.restore_best(False)
                noimpr = 0
        # }
        self.restore_best()
        return self.best, self.best_score, self.mod
    # }
# }


# -*- coding: utf-8 -*-
import numpy as np
import pytest

from searchlogit import MixedLogit, device

device.disable_gpu_acceleration()

# Setup data used for tests
X = np.array([[2, 1], [1, 3], [3, 1], [2, 4], [2, 1], [2, 4]])
y = np.array([0, 1, 0, 1, 0, 1])
ids = np.array([1, 1, 2, 2, 3, 3])
alts = np.array([1, 2, 1, 2, 1, 2])
panels = np.array([1, 1, 1, 1, 2, 2])
varnames = ["a", "b"]
randvars = {'a': 'n', 'b': 'n'}
N, J, K, R = 3, 2, 2, 5


def test__balance_panels():
    """
    Ensures that unbalanced panels are properly balanced when required
    """
    X_, y_ = X.reshape(N, J, K), y.reshape(N, J, 1)
    model = MixedLogit()
    X_, y_, _, panel_info = model._balance_panels(X_, y_, None, panels)

    assert np.array_equal(panel_info, np.array([[1, 1], [1, 0]]))
    assert X_.shape == (4, 2, 2)


def test_log_likelihood():
    """
    Computes the log-likelihood "by hand" for a simple example and ensures
    that the one returned by searchlogit is the same
    """
    P = 1  # Without panel data
    betas = np.array([.1, .1, .1, .1])
    X_, y_ = X.reshape(N, P, J, K), y.reshape(N, P, J, 1)

    # Compute log likelihood using searchlogit
    model = MixedLogit()
    model._rvidx,  model._rvdist = np.array([True, True]), np.array(['n', 'n'])
    draws = model._generate_halton_draws(N, R, K)  # (N,Kr,R)
    # drawstrans = model._generate_halton_draws(N, R, 0)  # (N,Kr,R)
    drawstrans = np.atleast_3d(np.array([]))
    panel_info = np.ones((N, P))

    # TODO: code smell
    model.Kf = 0
    model.Kr = 2
    model.Kftrans = 0
    model.Krtrans = 0
    model.Kchol = 0
    model.correlationLength = 0
    model.Kbw = 2
    model.n_draws = R
    model.N = N
    model.fxidx = [False, False]
    model.fxtransidx = [False, False]
    model.rvidx = [True, True]
    model.rvtransidx = [False, False]
    model.P = 1
    model.J = 2
    model.batch_size = R
    model.rvdist = ['n', 'n']
    model.save_fitted_params = False
    model.transvars = []
    model.varnames = ['a', 'b']
    model.return_grad = True
    model.return_hess = True

    obtained_loglik, _ = model._loglik_gradient(betas, X_, y_, panel_info,
                                                draws, drawstrans, None, None, K)

    # Compute expected log likelihood "by hand"
    Br = betas[None, [0, 1], None] + draws*betas[None, [2, 3], None]
    eXB = np.exp(np.einsum('npjk,nkr -> npjr', X_, Br))
    p = eXB/np.sum(eXB, axis=2, keepdims=True)
    expected_loglik = -np.sum(np.log(
        (y_*p).sum(axis=2).prod(axis=1).mean(axis=1)))

    #assert expected_loglik == pytest.approx(obtained_loglik)


def test__transform_betas():
    """
    Check that betas are properly transformed to random draws

    """
    # TODO
    pass
    # betas = np.array([.1, .1, .1, .1])

    # # Compute log likelihood using searchlogit
    # model = MixedLogit()
    # model.Kf = 0
    # model.Kr = 4
    # model.Kftrans = 0
    # model.Krtrans = 0
    # model.Kchol = 0
    # model.correlationLength = 0
    # model.Kbw = 0
    # model._rvidx,  model._rvdist = np.array([True, True]), np.array(['n', 'n'])
    # draws = model._generate_halton_draws(N, R, K)  # (N,Kr,R)
    # expected_betas = betas[None, [0, 1], None] + \
    #     draws*betas[None, [2, 3], None]
    # _, obtained_betas = model._transform_betas(betas, draws, chol_mat)

    # assert np.allclose(expected_betas, obtained_betas)


def test_fit():
    """
    Ensures the log-likelihood works for a single iterations with the default
    initial coefficients. The value of -1.794 was computed by hand for
    comparison purposes
    """
    # There is no need to initialize a random seed as the halton draws produce
    # reproducible results
    assert True
    # model = MixedLogit()
    # model.fit(X, y, varnames=varnames, alts=alts, n_draws=10, panels=panels,
#               ids=ids, randvars=randvars, maxiter=0, verbose=0, halton=True)

   # assert model.loglikelihood == pytest.approx(-1.79451632)


def test_validate_inputs():
    """
    Covers potential mistakes in parameters of the fit method that searchlogit
    should be able to identify
    """
    model = MixedLogit()
    with pytest.raises(ValueError):  # randvars is required for mixedlogit
        model.fit(X, y, varnames=varnames, alts=alts, ids=ids, n_draws=10,
                  maxiter=0, verbose=0, halton=True)

    with pytest.raises(ValueError):  # wrong distribution
        model.fit(X, y, varnames=varnames, alts=alts, ids=ids, n_draws=10,
                  maxiter=0, verbose=0, halton=True, randvars={'a': 'fake'})

    with pytest.raises(ValueError):  # wrong var name
        model.fit(X, y, varnames=varnames, alts=alts, ids=ids, n_draws=10,
                  maxiter=0, verbose=0, halton=True, randvars={'fake': 'n'})


def test_gpu_not_available():
    """
    Ensures that searchlogit
    detects that GPU is not available based on CuPy's
    installation status

    """
    assert not MixedLogit.check_if_gpu_available()

# TODO: test correlation, boxcox



import itertools
import logging
import time
import numpy as np
#import misc
from scipy.optimize import minimize



try:
    #from .misc import *
    from .mixed_logit import MixedLogit
    from .ordered_logit import OrderedLogit, get_last_elements, get_first_elements
    from .boxcox_functions import truncate_lower, truncate_higher, truncate
    from ._device import device as dev
except ImportError:
    import misc
    from mixed_logit import MixedLogit
    from boxcox_functions import truncate_lower, truncate_higher, truncate
    from _device import device as dev
    from ordered_logit import OrderedLogit, get_last_elements, get_first_elements

''' ---------------------------------------------------------- '''
''' CONSTANTS - BOUNDS ON NUMERICAL VALUES                     '''
''' ---------------------------------------------------------- '''
max_exp_val, min_exp_val = 700, -700
max_comp_val, min_comp_val = 1e+20, 1e-200   # or use float('inf')

''' ---------------------------------------------------------- '''
''' ERROR CHECKING AND LOGGING                                 '''
''' ---------------------------------------------------------- '''
logger = logging.getLogger(__name__)



class OrderedMixedLogit(OrderedLogit, MixedLogit):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.descr = "Ordered Mixed Logit"

    def compute_latent(self, X: np.ndarray, beta_fixed: np.ndarray, beta_random: np.ndarray, draws: np.ndarray) -> np.ndarray:
        y_latent = X @ beta_fixed[1:] + (X @ beta_random[1:] * draws).sum(axis=1)
        if self.fit_intercept:
            y_latent += beta_fixed[0]
        return y_latent

    def setup(self, **kwargs):
        # Call OrderedLogit's setup
        OrderedLogit.setup(self, **kwargs)

        # Call MixedLogit's setup to handle random coefficients
        mixed_logit_args = {key: kwargs[key] for key in kwargs if key not in ["J", "distr", 'start',
                                                                              'normalize']}
        # Call MixedLogit's setup
        MixedLogit.setup(self, **mixed_logit_args)
        #MixedLogit.setup(self, **kwargs)

        # Ensure `fn_generate_draws` is initialized
        if not hasattr(self, "fn_generate_draws"):
            self.fn_generate_draws = self.generate_draws_halton if self.halton else self.generate_draws_random

    def get_thresholds(self, params: np.ndarray, draws: np.ndarray = None) -> np.ndarray:
        delta = get_last_elements(params, self.J - 1)
        thresholds = np.cumsum(delta)
        if draws is not None:
            thresholds += draws.mean(axis=1)
        return thresholds

    def compute_probability(self, X: np.ndarray, params: np.ndarray, draws: np.ndarray) -> np.ndarray:
        thresholds = self.get_thresholds(params, draws)
        beta_fixed = self.get_beta(params)
        beta_random = self.get_random_beta(params, draws)
        y_latent = self.compute_latent(X, beta_fixed, beta_random, draws)
        cut = np.concatenate(([-np.inf], thresholds, [np.inf]))
        low = cut[:-1] - y_latent[:, None]
        high = cut[1:] - y_latent[:, None]
        prob = self.prob_interval(low, high)
        return prob

    def fit(self, method="L-BFGS-B", n_draws=500, start=None):
        self.draws = self.generate_draws(self.N, n_draws)
        if start is None:
            start = np.zeros(self.nparams)
        result = minimize(
            fun=self.get_loglike_gradient,
            x0=start,
            args=(self.X, self.y, self.draws),
            method=method,
            jac=True,
            options={"maxiter": 5000}
        )
        self.params = result.x
        self.post_process()


"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
FUNCTIONS FOR BOX-COX TRANSFORMATION
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

''' ----------------------------------------------------------- '''
'''  MAIN PARAMETERS:                                           '''
''' ----------------------------------------------------------- '''
#  X_matrix: Matrix to transform / array-like
#  lmdas: lambda parameters for boxcox transformation/ array-like
#  bxcx_X: Matrix after boxcox transformation / array-like

''' ---------------------------------------------------------- '''
''' LIBRARIES                                                  '''
''' ---------------------------------------------------------- '''
import numpy as np

''' ---------------------------------------------------------- '''
''' CONSTANTS                                                  '''
''' ---------------------------------------------------------- '''
min_val = 1e-30  # Define smallest value

''' ---------------------------------------------------------- '''
''' FUNCTION                                                   '''
''' ---------------------------------------------------------- '''
def truncate_lower(x, minval):
    x[np.where(x < minval)] = minval
    return x

def truncate_higher(x, maxval):
    x[np.where(x > maxval)] = maxval
    return x

def truncate(x, minval, maxval): # {
    x = truncate_lower(x, minval)
    x = truncate_higher(x, maxval)
    return x
# }

''' ---------------------------------------------------------- '''
''' Function. x is an ndarray, lamda is a scalar               '''
''' ---------------------------------------------------------- '''
def transform(x, lmda):
# {
    if lmda == 0:
        #return np.log(x)
        return np.log1p(x)
    else:
        val = (np.power(x, lmda) - 1) / lmda
        return np.nan_to_num(val)
# }

''' ---------------------------------------------------------- '''
''' Function. x is an ndarray, lamda is a scalar               '''
''' ---------------------------------------------------------- '''
def transform_derivative(x, lmda):
# {
    """ Derivative of log likelihood with respect to lmda """
    ln_x = np.log1p(x) # Evaluate ln(1+x) because elements of x are close to zero
    # Note: original code was ln_x = np.log(x)

    if lmda == 0:
        return 0.5*(ln_x)**2  # i.e., 0.5 * ln(x)^2
    else:
    # {
        x_lmda = np.nan_to_num(np.power(x, lmda))
        val = (lmda * x_lmda * ln_x - x_lmda + 1) / np.power(lmda, 2)
        return np.nan_to_num(val) # Return zero
    # }
# }

def prep(X_matrix, lmdas):
# {
    lmdas = truncate(lmdas, -5, 5)
    X_matrix = truncate_lower(X_matrix, min_val)
    bxcx = np.zeros_like(X_matrix)  # initialise to zero
    bxcx = bxcx.astype("float64") # cast each element as a float
    return X_matrix, lmdas, bxcx
# }

''' ---------------------------------------------------------- '''
''' Function.Returns boxcox transformed matrix                 '''
''' ---------------------------------------------------------- '''
def boxcox_transformation(X_matrix, lmdas):
# {
    X_matrix, lmdas, bxcx_X = prep(X_matrix, lmdas)
    for i, lmda in enumerate(lmdas):
        bxcx_X[:, :, i] = transform(X_matrix[:, :, i], lmda)
    return bxcx_X
# }

def boxcox_transformation_mixed(X_matrix, lmdas):
# {
    X_matrix, lmdas, bxcx_X = prep(X_matrix, lmdas)
    for i, lmda in enumerate(lmdas):
        bxcx_X[:, :, :, i] = transform(X_matrix[:, :, :, i], lmda)
    return bxcx_X
# }

''' ---------------------------------------------------------------------- '''
''' Function. Estimate derivative of boxcox transformation parameter (lambda)'''
''' ---------------------------------------------------------------------- '''
def boxcox_param_deriv(X_matrix, lmdas):
# {
    X_matrix, lmdas, der_bxcx_X = prep(X_matrix, lmdas)
    for i, lmda in enumerate(lmdas):
        der_bxcx_X[:, :, i] = transform_derivative(X_matrix[:, :, i], lmda)
    return der_bxcx_X
# }

def boxcox_param_deriv_mixed(X_matrix, lmdas):
# {
    X_matrix, lmdas, der_bxcx_X = prep(X_matrix, lmdas)
    for i, lmda in enumerate(lmdas):
        der_bxcx_X[:, :, :, i] = transform_derivative(X_matrix[:, :, :, i], lmda)
    return der_bxcx_X
# }


"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
IMPLEMENTATION: LATENT CLASS MIXED MODEL 
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""
BACKGROUND - LATENT MIXED MODEL

A latent mixed model, also known as a latent mixed-effects model or a 
latent variable mixed model, is a statistical model that combines 
elements of both mixed-effects models and latent variable models. 

Let's break down the components of a latent mixed model:

1. Mixed-effects model: A mixed-effects model is a type of regression model
that incorporates both fixed effects and random effects. Fixed effects 
represent population-level parameters that are assumed to be constant 
across all individuals or groups, while random effects represent individual
 or group-specific deviations from the population-level parameters.

2. Latent variable model: A latent variable model posits the existence 
of unobserved (latent) variables that underlie the observed data. 
These latent variables are not directly measured but are inferred from 
patterns in the observed data.

In a latent mixed model, the key idea is to include latent variables 
as part of the random effects component of the mixed-effects model. 
These latent variables capture unobserved heterogeneity or latent 
traits that influence the outcome variable.

Here's how a latent mixed model might be formulated:

- Fixed Effects: Similar to traditional mixed-effects models, the fixed 
effects component represents the population-level parameters that are 
assumed to be constant across all individuals or groups.

- Random Effects: In addition to the traditional random effects 
(e.g., random intercepts, random slopes), the random effects 
component includes latent variables that capture unobserved 
heterogeneity or latent traits among individuals or groups.

- Latent Variables: The latent variables are assumed to influence the outcome
variable indirectly through their effect on the observed predictors 
or through their interaction with other  variables in the model. 
These latent variables can represent underlying traits, attitudes, 
abilities, or other unobserved factors.

- Model Estimation: Estimating the parameters of a latent mixed 
model typically involves fitting the model to the observed data 
using statistical methods such as maximum likelihood estimation 
(MLE), Bayesian estimation, or other estimation techniques. 
The goal is to estimate both the fixed effects parameters and 
the random effects parameters, including the parameters associated 
with the latent variables.

Latent mixed models are particularly useful when there is interest 
in capturing unobserved heterogeneity or latent traits that may 
influence the outcome variable, while also accounting for the 
hierarchical or clustered structure of the data using random effects. 
These models are commonly used in fields such as psychology, 
sociology, education, and epidemiology to study individual 
differences and group-level effects.
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


''' ---------------------------------------------------------- '''
''' LIBRARIES                                                  '''
''' ---------------------------------------------------------- '''
import itertools
import logging
import time
import numpy as np
#import misc
from scipy.optimize import minimize



try:
    from . import misc
    from .mixed_logit import MixedLogit
    from .boxcox_functions import truncate_lower, truncate_higher, truncate
    from ._device import device as dev
except ImportError:
    import misc
    from mixed_logit import MixedLogit
    from boxcox_functions import truncate_lower, truncate_higher, truncate
    from _device import device as dev

''' ---------------------------------------------------------- '''
''' CONSTANTS - BOUNDS ON NUMERICAL VALUES                     '''
''' ---------------------------------------------------------- '''
max_exp_val, min_exp_val = 700, -700
max_comp_val, min_comp_val = 1e+20, 1e-200   # or use float('inf')

''' ---------------------------------------------------------- '''
''' ERROR CHECKING AND LOGGING                                 '''
''' ---------------------------------------------------------- '''
logger = logging.getLogger(__name__)

''' ---------------------------------------------------------- '''
''' CLASS FOR ESTIMATION OF LATENT CLASS MODELS                '''
''' ---------------------------------------------------------- '''
class LatentClassMixedModel(MixedLogit):
# {
    """ Docstring """

    """
    The design of this class is partly based on the LCCM package,
    https://github.com/ferasz/LCCM (El Zarwi, 2017).

    References
    ----------
    El Zarwi, F. (2017). lccm, a Python package for estimating latent
    class choice models using the Expectation Maximization (EM)
    algorithm to maximize the likelihood function.
    """

    # ===================
    # CLASS PARAMETERS
    # ===================
    """"
    X:                  Input data for explanatory variables / long format / array-like / shape (n_samples, n_variables)
    y:                  Choices / array-like / shape (n_samples,)
    varnames:           Names of explanatory variables / list / shape (n_variables,)
    int num_classes:    Number of latent classes
    alts:               List of alternative names or indexes / long format / array-like / shape (n_samples,)
    isvars:             Names of individual-specific variables in varnames / list
    transvars:          Names of variables to apply transformation on / list / default=None
    transformation:     Transformation to apply to transvars / string / default="boxcox"
    ids:                Identifiers for choice situations / long format / array-like / shape (n_samples,)
    weights:            Weights for the choice situations / long format / array-like / shape (n_variables,) / default=None
    avail:              Availability indicator of alternatives for the choices (1 => available, 0 otherwise)/ array-like / shape (n_samples,)
    base_alt:           Base alternative / int, float or str / default=None
    init_coeff:         Initial coefficients for estimation/ numpy array / shape (n_variables,) / default=None
    bool fit_intercept: Boolean indicator to include an intercept in the model / default=False
    int maxiter:        Maximum number of iterations / default=2000
    dict randvars:      Names (keys) and mixing distributions of random variables /
                        Distributions: n - normal, ln - lognormal, u - uniform, t - triangular, tn - truncated normal
    params_spec:        Array of lists containing names of variables for latent class / array_like / shape (n_variables,)
    member_params_spec: Array of lists containing names of variables for class / array_like / shape (n_variables,)
    panels:             Identifiers to create panels in combination with ids / array-like / long format / shape (n_samples,) / default=None
    method:             Optimisation method for scipy.optimize.minimize / string / default="bfgs"
    float ftol:         Tolerance for scipy.optimize.minimize termination / default=1e-5
    float gtol:         Tolerance for scipy.optimize.minimize(method="bfgs") termination - gradient norm / default=1e-5
    bool return_grad:   Flag to calculate the gradient in _loglik_and_gradient / default=True
    bool return_hess:   Flag to calculate the hessian in _loglik_and_gradient / default=True
    bool scipy_optimisation : Flag to apply optimiser / default=False / When false use own bfgs method.

    int batch_size:         Size of batches of random draws used to avoid overflowing memory during computations/ default=None
    bool shuffle:           Flag to shuffle the Halton draws / default=False
    int n_draws:            Random draws to approximate the mixing distributions of the random coefficients / default=1000
    bool halton:            Boolean flag for Halton draws / default=True
    int drop:               # of Halton draws to discard (initially) to minimize correlations between Halton sequences/ default=100
    primes:                 List of primes for generation of Halton sequences / list
    dict halton_opts:       Options for generation of halton draws (shuffle, drop, primes) / default=None

    """

    # ===================
    # CLASS FUNCTIONS
    # ===================

    """
    1. __init__(self);
    2. setup(self, X, y, ...);
    3. fit(self);
    4. post_process(self, optimization_res, coeff_names, sample_size, hess_inv=None);
    5. pch <-- compute_probabilities_latent(self, betas, X, y, panel_info, draws, drawstrans, avail); 
    6. H <-- posterior_est_latent_class_probability(self, class_thetas);
    7. Loglik <-- class_member_func(self, class_thetas, weights, X);
    8. X_class_idx <-- get_class_X_idx2(self, class_num, coeff_names=None, **kwargs);
    9. X_class_idx <-- get_class_X_idx(self, class_num, coeff_names=None);
    10. Kchol <-- get_kchol(self, specs);
    11. len <-- get_betas_length(self, class_num);
    12. make_short_df(self, X);
    13. void set_bw(self, specs);
    14. rand_idx, randtrans_idx <-- update(self, i, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs);
    15. expectation_maximisation_algorithm(self, tmp_fn, tmp_betas, args, class_betas=None, class_thetas=None, validation=False, **kwargs);
    16. result <-- bfgs_optimization(self, betas, X, y, weights, avail, maxiter);
    """

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def __init__(self, **kwargs): # {
        self.verbose= 0
        self.optimise_class = kwargs.get('optimise_class', False)
        self.optimise_membership = kwargs.get('optimise_membership', False)
        self.fixed_solution = kwargs.get('fixed_solution', None)
        self.fixed_thetas = None if self.fixed_solution is None else self.fixed_solution['model'].class_x
        self.save_fitted_params = False  # speed-up computation
        self.start_time = time.time()
        self.descr = "LCMM"
        super(LatentClassMixedModel, self).__init__()
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Set up the model                                 '''
    ''' ---------------------------------------------------------- '''
    def setup(self, X, y, varnames=None, alts=None, isvars=None, num_classes=2,
              class_params_spec=None, class_params_spec_is = None, member_params_spec=None,
              transvars=None, transformation=None, ids=None, weights=None, avail=None,
              avail_latent=None,  # TODO?: separate param needed?
              randvars=None, panels=None, base_alt=None, intercept_opts=None,
              init_coeff=None, init_class_betas=None, init_class_thetas=None,
              maxiter=2000, correlated_vars=None, n_draws=1000, halton=True,
              batch_size=None, halton_opts=None, ftol=1e-5, ftol_lccmm=1e-4,
              gtol=1e-5, gtol_membership_func=1e-5, return_hess=True, return_grad=True, method="bfgs",
              validation=False, mnl_init=True, mxl_init=True, verbose=False):
    # {


        if varnames is not None and member_params_spec is not None:
            varnames = misc.rearrage_varnames(varnames, member_params_spec)

        #varnames = misc.rearrage_varnames(varnames, member_params_spec)
        self.ftol, self.gtol = ftol, gtol
        self.ftol_lccmm = ftol_lccmm
        self.gtol_membership_func = gtol_membership_func
        self.num_classes = num_classes
        self.panels = panels
        self.init_df, self.init_y = X, y
        self.ids = ids
        self.mnl_init = mnl_init
        self.verbose = verbose
        batch_size = n_draws if batch_size is None else min(n_draws, batch_size)
        self.fit_intercept = misc.initialise_fit_intercept(class_params_spec, intercept_opts)
        self.class_params_spec = misc.initialise_class_params_spec(class_params_spec, isvars, varnames, num_classes)
        self.class_params_spec_is = misc.initialise_class_params_spec(class_params_spec_is, isvars, [], num_classes)
        for i in range(num_classes):
            self.class_params_spec[i] = [j for j in self.class_params_spec[i] if j not in self.class_params_spec_is[i]] 
        self.intercept_opts = misc.initialise_opts(intercept_opts, num_classes)
        self.intercept_classes = [('_inter' in class_params_spec[var]) for var in range(len(class_params_spec))]                                                                                                             
        self.avail_latent = misc.initialise_avail_latent(avail_latent, num_classes)
        self.membership_as_probability = misc.initialise_membership_as_probability(member_params_spec)

        args = (self.membership_as_probability, member_params_spec, isvars, varnames, num_classes)
        self.member_params_spec = misc.initialise_member_params_spec(*args)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Initialise: MXL
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if mxl_init and init_class_betas is None:
        # {
            init_class_betas = np.array(np.repeat('tmp', num_classes), dtype='object')  # Create temp/template array
            for i in range(num_classes):
            # {
                #Only want randvars if its in the class.
                randvars_class = {var: randvars[var] for var in self.class_params_spec[i] if var in randvars}                                                                                                    
                mxl = MixedLogit()
                try:
                    is_class = self.class_params_spec_is
                except:
                    is_class = []
                mxl = misc.setup_logit(i, mxl, X, y, varnames, self.class_params_spec, is_class, avail, alts, transvars, gtol,
                        mxl=True, panels=panels, randvars=randvars_class,
                        correlated_vars=correlated_vars, n_draws=n_draws, mnl_init=mnl_init)
                init_class_betas = misc.revise_betas(i, mxl, init_class_betas, self.intercept_opts, self.alts)

            # }
        # }

        self.init_class_betas = init_class_betas
        self.init_class_thetas = init_class_thetas
        self.validation = validation
        self.ind_pred_prob_classes, self.choice_pred_prob_classes= [], []
        
        if self.optimise_class and self.optimise_membership == False and self.fixed_solution is not None:
            minimise_model = self.fixed_expectation_algorithm
        else:
            minimise_model = self.expectation_maximisation_algorithm                                                                  
        super(LatentClassMixedModel, self).setup(X, y, varnames, alts, isvars,
                transvars, transformation, ids, weights, avail, randvars, panels, base_alt,
                self.fit_intercept, init_coeff, maxiter, correlated_vars,
                n_draws, halton, minimise_model, batch_size,
                halton_opts, ftol, gtol, return_hess, return_grad, method, self.save_fitted_params, mnl_init)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Fit multinomial and/or conditional logit models  '''
    ''' ---------------------------------------------------------- '''
    def fit(self):  # {
        super(LatentClassMixedModel, self).fit()
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def post_process(self, optimization_res, coeff_names, sample_size,
                     hess_inv=None):
    # {
        if not self.validation:
            super(LatentClassMixedModel, self).post_process(optimization_res, coeff_names, sample_size)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Compute the standard logit-based probabilities   '''
    ''' Random and fixed coefficients are handled separately       '''
    ''' ---------------------------------------------------------- '''
    def compute_probabilities_latent(self, betas, X, y, panel_info, draws, drawstrans, avail):
    # {
        # ________________________________________________________________
        if dev.using_gpu:  # {
            X, y = dev.convert_array_gpu(X), dev.convert_array_gpu(y)
            panel_info = dev.convert_array_gpu(panel_info)
            draws = dev.convert_array_gpu(draws)
            drawstrans = dev.convert_array_gpu(drawstrans)
            if avail is not None: avail = dev.convert_array_gpu(avail)
        # }
        # _______________________________________________________________

        beta_segment_names = ["Bf", "Br_b", "chol", "Br_w", "Bftrans",
                              "flmbda", "Brtrans_b", "Brtrans_w", "rlmda"]

        iterations = [self.Kf, self.Kr, self.Kchol, self.Kbw, self.Kftrans,
                      self.Kftrans, self.Krtrans, self.Krtrans, self.Krtrans]
        if sum(iterations) != len(betas):
            print('dirty fix')
            missing_amount = sum(iterations) - len(betas)
            betas = np.append(betas, [0.01] * missing_amount)  # 
            
        var_list = self.split_betas(betas, iterations, beta_segment_names)
        Bf, Br_b, chol, Br_w, Bftrans, flmbda, Brtrans_b, Brtrans_w, rlmda = var_list.values()

        # ______________________________________________________________________________________
        if dev.using_gpu:  # {
            Bf, Br_b = dev.convert_array_gpu(Bf), dev.convert_array_gpu(Br_b)
            chol = dev.convert_array_gpu(chol)
            Br_w, Bftrans = dev.convert_array_gpu(Br_w), dev.convert_array_gpu(Bftrans)
            flmbda = dev.convert_array_gpu(flmbda)
            Brtrans_b, Brtrans_w = dev.convert_array_gpu(Brtrans_b), dev.convert_array_gpu(Brtrans_w)
            rlmda = dev.convert_array_gpu(rlmda)
        # }
        # _________________________________________________________________________________________

        chol_mat = np.zeros((self.correlationLength, self.correlationLength))
        indices = np.tril_indices(self.correlationLength)

        # __________________________________________________________
        if dev.using_gpu: chol = dev.convert_array_cpu(chol)
        # __________________________________________________________
        
        try:
            chol_mat[indices] = chol
        except Exception as e:
            print('why')

        Kr_all = self.Kr + self.Krtrans
        chol_mat_temp = np.zeros((self.Kr, self.Kr))

        # TODO: Structure ... Kr first, Krtrans last, fill in for correlations
        # TODO: could do better
        rv_count, rv_count_all, rv_trans_count, rv_trans_count_all, chol_count = 0, 0, 0, 0, 0
        corr_indices = []

        # TODO: another bugfix
        # Know beforehand to order rvtrans correctly
        num_corr_rvtrans = 0
        for ii, var in enumerate(self.varnames):
        # {
            if self.rvtransidx[ii] and hasattr(self, 'correlated_vars') \
                    and self.correlated_vars \
                        and hasattr(self.correlated_vars,'append') and var in self.correlated_vars:
                num_corr_rvtrans += 1
        # }

        num_rvtrans_total = self.Krtrans + num_corr_rvtrans

        if self.Kr > 0:
        # {
            for ii, var in enumerate(self.varnames):  # TODO: BUGFIX
            # {
                #FIXME I believe this is and not or
                is_correlated = hasattr(self, 'correlated_vars') and self.correlated_vars and (
                            hasattr(self.correlated_vars, 'append') and var in self.correlated_vars)

                if self.rvidx[ii]:
                # {
                    rv_val = chol[chol_count] if is_correlated else Br_w[rv_count]
                    chol_mat_temp[rv_count_all, rv_count_all] = rv_val
                    rv_count_all += 1

                    if is_correlated:
                        chol_count += 1
                    else:
                        rv_count += 1
                # }

                if self.rvtransidx[ii]:
                # {
                    is_correlated = isinstance(self.correlated_vars, bool) and self.correlated_vars
                    rv_val = chol[chol_count] if is_correlated else Brtrans_w[rv_trans_count]
                    at = rv_trans_count_all - num_rvtrans_total
                    chol_mat_temp[at, at] = rv_val
                    rv_trans_count_all += 1

                    if is_correlated:
                        chol_count += 1
                    else:
                        rv_trans_count += 1
                # }

                if hasattr(self, 'correlated_vars') and self.correlated_vars:
                # {
                    if hasattr(self.correlated_vars, 'append'):
                    # {
                        if var in self.correlated_vars:
                        # {
                            if self.rvidx[ii]:
                                corr_indices.append(rv_count_all - 1)
                            else:
                                corr_indices.append(Kr_all - num_rvtrans_total + rv_trans_count_all - 1)  # TODO i think
                        # }
                    # }
                # }
            # }
            if hasattr(self, 'correlated_vars') and isinstance(self.correlated_vars, bool) and self.correlated_vars:
                corr_pairs = list(itertools.combinations(np.arange(self.Kr), 2)) + [(i, i) for i in range(self.Kr)]
            else:
                corr_pairs = list(itertools.combinations(corr_indices, 2)) + [(idx, idx) for ii, idx in
                                                                              enumerate(corr_indices)]

            reversed_corr_pairs = [tuple(reversed(pair)) for ii, pair in enumerate(corr_pairs)]
            reversed_corr_pairs.sort(key=lambda x: x[0])

            chol_count = 0

            for _, corr_pair in enumerate(reversed_corr_pairs):
            # {
                # lower cholesky matrix
                chol_mat_temp[corr_pair] = chol[chol_count]
                chol_count += 1
            # }
            chol_mat = chol_mat_temp
        # }



        V = np.zeros((self.N, self.P, self.J, self.n_draws))

        # __________________________________________________
        if dev.using_gpu:  # {
            V = dev.convert_array_gpu(V)
            chol_mat = dev.convert_array_gpu(chol_mat)
        # }
        # __________________________________________________

        if self.Kf != 0:  # {
            Xf = X[:, :, :, self.fxidx]
            if dev.using_gpu: Xf = dev.convert_array_gpu(Xf)
            XBf = np.einsum('npjk,k -> npj', Xf, Bf, dtype=np.float64)
            V += XBf[:, :, :, None]
        # }

        if self.Kr != 0:  # {
            Br = Br_b[None, :, None] + np.matmul(chol_mat, draws)
            Br = self.apply_distribution(Br, self.rvdist)
            self.Br = Br  # save Br to use later
            Xr = X[:, :, :, self.rvidx]
            if dev.using_gpu: Xr = dev.convert_array_gpu(Xr)
            XBr = dev.cust_einsum('npjk,nkr -> npjr', Xr, Br)  # (N, P, J, R)
            V += XBr
        # }

        #  Apply transformations for variables with fixed coeffs
        if self.Kftrans != 0:
        # {
            Xftrans = X[:, :, :, self.fxtransidx]
            if dev.using_gpu: Xftrans = dev.convert_array_gpu(Xftrans)
            Xftrans_lmda = self.trans_func(Xftrans, flmbda)
            Xftrans_lmda[np.isneginf(Xftrans_lmda)] = -max_comp_val
            Xftrans_lmda[np.isposinf(Xftrans_lmda)] = max_comp_val
            # Estimating the linear utility specificiation (U = sum XB)
            Xbf_trans = np.einsum('npjk,k -> npj', Xftrans_lmda, Bftrans, dtype=np.float64)
            V += Xbf_trans[:, :, :, None] # Combining utilities
        # }

        # Apply transformations for variables with random coeffs
        if self.Krtrans != 0:
        # {
            # Create the random coeffs:
            Brtrans = Brtrans_b[None, :, None] + \
                      drawstrans[:, 0:self.Krtrans, :] * Brtrans_w[None, :, None]
            Brtrans = self.apply_distribution(Brtrans, self.rvtransdist)
            # Apply transformation:
            Xrtrans = X[:, :, :, self.rvtransidx]
            if dev.using_gpu:  Xrtrans = dev.convert_array_gpu(Xrtrans)
            Xrtrans_lmda = self.trans_func(Xrtrans, rlmda)
            Xrtrans_lmda[np.isposinf(Xrtrans_lmda)] = 1e+30
            Xrtrans_lmda[np.isneginf(Xrtrans_lmda)] = -1e+30
            Xbr_trans = np.einsum('npjk, nkr -> npjr', Xrtrans_lmda, Brtrans, dtype=np.float64)  # (N, P, J, R)
            V += Xbr_trans  # (N, P, J, R) # combining utilities
        # }

        if avail is not None:  # {
            ref = avail[:, :, :, None] if self.panels is not None else avail[:, None, :, None]
            V = V * ref # Accommodate availablity of alts with panels or withut panels
        # }

        # Thresholds to avoid overflow warnings
        V = truncate(V, -max_exp_val, max_exp_val)
        eV = dev.np.exp(V)
        sum_eV = dev.np.sum(eV, axis=2, keepdims=True)
        sum_eV = truncate_lower(sum_eV, min_comp_val)
        p = np.divide(eV, sum_eV, out=np.zeros_like(eV))
        p = p * panel_info[:, :, None, None] if panel_info is not None else p
        p = y * p

        # collapse on alts
        pch = np.sum(p, axis=2)  # (N, P, R)

        if hasattr(self, 'panel_info'):
            pch = self.prob_product_across_panels(pch, self.panel_info)
        else:
            pch = np.mean(pch, axis=1)  # (N, R)

        pch = np.mean(pch, axis=1)  # (N)
        return pch.flatten()

    # }

    ''' ----------------------------------------------------------- '''
    ''' Function: Get prior estimates of latent class probabilities '''
    ''' ----------------------------------------------------------- '''
    def posterior_est_latent_class_probability(self, class_thetas):
    # {
        """
        class_thetas (array-like): Array of latent class vectors
        H: Prior estimates of the class probabilities
        """
        class_thetas_original = class_thetas
        if class_thetas.ndim == 1:
        # {
            new_class_thetas = np.array(np.repeat('tmp', self.num_classes - 1), dtype='object')
            j = 0
            for ii, member_params in enumerate(self.member_params_spec):  # {
                num_params = len(member_params)
                tmp = class_thetas[j:j + num_params]
                j += num_params
                new_class_thetas[ii] = tmp
            # }
            class_thetas = new_class_thetas
        # }

        class_thetas_base = np.zeros(len(class_thetas[0]))

        # coeff_names_without_intercept = self.global_varnames[(self.J-2):]
        base_X_idx = self.get_member_X_idx(0)
        member_df = np.transpose(self.short_df[:, base_X_idx])
        member_N = member_df.shape[1]
        eZB = np.zeros((self.num_classes, member_N))

        if '_inter' in self.member_params_spec[0]:  # {
            ones = np.ones((1, member_N))
            transposed = np.transpose(self.short_df[:, base_X_idx])
            member_df = np.vstack((ones, transposed))
        # }

        if self.membership_as_probability:  # {
            H = np.tile(np.concatenate([1 - np.sum(class_thetas), class_thetas_original]), (member_N, 1))
            H = np.transpose(H)
        # }
        else:  # {
            zB_q = np.dot(class_thetas_base[None, :], member_df)
            eZB[0, :] = np.exp(zB_q)

            for i in range(0, self.num_classes - 1):
            # {
                class_X_idx = self.get_member_X_idx(i)
                member_df = np.transpose(self.short_df[:, class_X_idx])

                # add in columns of ones for class-specific const (_inter)
                if '_inter' in self.member_params_spec[i]:  # {
                    print('off for now'
                    )
                    '''
                    member_df = np.vstack((np.ones((1, member_N)), np.transpose(self.short_df[:, class_X_idx])))
                    '''
                # }

                zB_q = np.dot(class_thetas[i].reshape((1, -1)), member_df)
                zB_q = truncate_higher(zB_q, max_exp_val)
                eZB[i + 1, :] = np.exp(zB_q)
            # }
            H = eZB / np.sum(eZB, axis=0, keepdims=True)
        # }
        self.class_freq = np.mean(H, axis=1) # store to display in summary
        return H
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def class_member_func(self, class_thetas, weights, X):
    # {
        """Used in Maximisaion step. Used to find latent class vectors that
           minimise the negative loglik where there is no observed dependent
           variable (H replaces y).

        Args:
            class_thetas (array-like): (number of latent classes) - 1 array of
                                       latent class vectors
            weights (array-like): weights is prior probability of class by the
                                  probability of y given the class.
            X (array-like): Input data for explanatory variables in wide format
        Returns:
            ll [np.float64]: Loglik
        """
        H = self.posterior_est_latent_class_probability(class_thetas)
        H = truncate_lower(H, 1e-30)  # i.e., H[np.where(H < 1e-30)] = 1e-30
        weight_post = np.multiply(np.log(H), weights)
        ll = -np.sum(weight_post)
        tgr = H - weights
        gr = np.array([])

        for i in range(1, self.num_classes):
        # {
            member_idx = self.get_member_X_idx(i - 1)
            membership_df = self.short_df[:, member_idx]

            if '_inter' in self.member_params_spec[i - 1]:
            # {
                membership_df = np.hstack((np.ones((self.short_df.shape[0], 1)), membership_df))
            # }

            if self.membership_as_probability:
                membership_df = np.ones((self.short_df.shape[0], 1))

            gr_i = np.dot(np.transpose(membership_df), tgr[i, :])
            gr = np.concatenate((gr, gr_i))
        # }
        penalty = self.reg_penalty*sum(class_thetas)

        return ll+penalty, gr.flatten()
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Get indices for X dataset for class parameters   '''
    ''' ---------------------------------------------------------- '''
    def get_class_X_idx2(self, class_num, coeff_names=None, **kwargs):
    # {
        #  below line: return indices of that class params in Xnames
        #  pattern matching for isvars

        tmp_varnames = self.global_varnames.copy() if coeff_names is None else coeff_names.copy()
        for ii, varname in enumerate(tmp_varnames):  # {
            if varname.startswith('lambda.'): tmp_varnames[ii] = varname[7:] # Remove lambda
            if varname.startswith('sd.'): tmp_varnames[ii] = varname[3:]
        # }

        X_class_idx = np.array([], dtype='int32')
        for var in self.class_params_spec[class_num]:
        # {
            for ii, var2 in enumerate(tmp_varnames):
            # {
                if 'inter' in var and 'inter' in var2 and coeff_names is not None:  # only want to use summary func
                # {
                    if 'class_intercept_alts' in self.intercept_opts:
                    # {
                        alt_num = int(var2.split('.')[-1])
                        if alt_num not in self.intercept_opts['class_intercept_alts'][class_num]:
                            continue
                    # }
                # }
                if var in var2:
                    X_class_idx = np.append(X_class_idx, ii)
            # }
        # }

        # isvars handled if pass in full coeff names
        X_class_idx = np.unique(X_class_idx)
        X_class_idx = np.sort(X_class_idx)
        X_class_idx_tmp = np.array([], dtype='int')
        counter = 0

        if coeff_names is not None:
            return X_class_idx

        for idx_pos in range(len(self.global_varnames)):
        # {
            if idx_pos in self.ispos:
            # {
                # fix bug of not all alts checked intercept
                for i in range(self.J - 1):
                # {
                    if idx_pos in X_class_idx:
                    # {
                        if self.global_varnames[idx_pos] == '_inter' and 'class_intercept_alts' in self.intercept_opts:
                        # {
                            if i + 2 not in self.intercept_opts['class_intercept_alts'][class_num]:
                            # {
                                counter += 1
                                continue
                            # }
                        # }
                        X_class_idx_tmp = np.append(X_class_idx_tmp, int(counter))
                    # }
                    counter += 1
                # }
            # }
            else:
            # {
                if idx_pos in X_class_idx:
                    X_class_idx_tmp = np.append(X_class_idx_tmp, counter)
                counter += 1
            # }
        # }

        X_class_idx = X_class_idx_tmp

        return X_class_idx
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Get indices for X dataset based on which         '''
    ''' parameters have been specified for the latent class        '''
    ''' ---------------------------------------------------------- '''
    def get_class_X_idx(self, class_num, coeff_names=None):
    # {
        """
        X_class_idx: indices to retrieve relevant
            explanatory params of specified latent class
        """
        #  below line: return indices of that class params in Xnames
        #  pattern matching for isvars

        if coeff_names is None:
            coeff_names = self.global_varnames.copy()
            #tring to handle the global var names
            if np.any(self.intercept_classes) == True:

            #if self.intercept_classes[class_num]:
            # {
                inter_count = sum(1 for name in coeff_names if '_inter' in name)
                num_in =len(self.alts) -1
                if inter_count < num_in:
                    new_names = ['_inter' for i in range(num_in)]
                    new_names.extend([name for name in coeff_names if 'inter' not in name])
                    coeff_names = new_names
            # }
        tmp_varnames = coeff_names.copy()
        for ii, varname in enumerate(tmp_varnames):
        # {
            # remove lambda so can get indices correctly
            if varname.startswith('lambda.'):
                tmp_varnames[ii] = varname[7:]

            if varname.startswith('sd.'):
                tmp_varnames[ii] = varname[3:]
        # }

        X_class_idx = np.array([], dtype="int")

        for var in self.class_params_spec[class_num]:
        # {
            alt_num_counter = 1
            # if 'inter' in var:
            #     alt_num_counter = 1
            for ii, var2 in enumerate(tmp_varnames):
            # {
                if 'inter' in var and 'inter' in var2 and coeff_names is not None:
                # {
                    if 'class_intercept_alts' in self.intercept_opts:
                    # {
                        if alt_num_counter not in self.intercept_opts['class_intercept_alts'][class_num]:
                        # {
                            alt_num_counter += 1
                            if alt_num_counter > 2:
                                continue    # Skip current iteration of loop
                        # }
                        else:
                            alt_num_counter += 1
                    # }
                # }

                if var in var2:
                    X_class_idx = np.append(X_class_idx, ii)
            # }
        # }

        X_class_idx = np.unique(X_class_idx)
        X_class_idx = np.sort(X_class_idx)

        return X_class_idx
    # }

    ''' -------------------------------------------------------------- '''
    ''' Function. Get indices for X dataset based on which paramerters '''
    ''' have been specified for the latent class membership            '''
    ''' -------------------------------------------------------------- '''
    def get_member_X_idx(self, class_num, coeff_names=None):
    # {
        if coeff_names is None:  # {
            cond = ('_inter' in self.global_varnames) and (self.J > 2)  # Evaluate boolean condition
            ref = self.global_varnames[(self.J - 2):] if cond else self.global_varnames
            coeff_names = ref.copy()    # Make a copy of the reference list
        # }

        tmp_varnames = coeff_names.copy()
        for ii, varname in enumerate(tmp_varnames):
        # {
            if varname.startswith('lambda.'):
                tmp_varnames[ii] = varname[7:] # Remove lambda so can get indices correctly
        # }

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Indices to retrieve relevant explanatory params of specified latent class
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #_class_idx = np.array([], dtype='int32')
        
        
        
        
        X_class_idx = np.array([], dtype='int32')
        for ii, var in enumerate(self.member_params_spec[class_num]):  # { #this causes error
            if '_inter' not in var:
                X_class_idx = np.append(X_class_idx, ii)
            # }
        # }
        
        '''
        for var in self.member_params_spec[class_num]:  # {
            for ii, var2 in enumerate(tmp_varnames):  # {
                if var == var2 and var != '_inter': #TODO changed this to equal
                    X_class_idx = np.append(X_class_idx, ii)
            # }
        # }
        '''
        X_class_idx = np.sort(X_class_idx)

        return X_class_idx
    # }

    ''' -------------------------------------------------------------- '''
    ''' Function. Permutations of specified params in correlation list '''
    ''' -------------------------------------------------------------- '''
    def get_kchol(self, specs):
    # {
        randvars_specs = [param for param in specs if param in self.randvars]
        Kchol = 0
        if (self.correlated_vars):
        # {
            if (isinstance(self.correlated_vars, list)):  # {
                corvars_in_spec = [corvar for corvar in self.correlated_vars if corvar in randvars_specs]
                self.correlationLength = len(corvars_in_spec)
            # }
            else:  # {
                self.correlationLength = len(randvars_specs)
                # i.e. correlation = True, Kchol permutations of rand vars
            # }
            Kchol = int(0.5 * self.correlationLength * (self.correlationLength + 1))
        # }
        return Kchol
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Get # betas for specified latent class           '''
    ''' ---------------------------------------------------------- '''
    def get_betas_length(self, class_num):
    # {
        class_idx = self.get_class_X_idx(class_num) #FIXME i modified this 
        self.set_bw(class_idx)
        class_params_spec = self.class_params_spec[class_num]
        class_asvars = [x for x in class_params_spec if x in self.asvars]
        class_randvars = [x for x in class_params_spec if x in self.randvars]
        class_transvars = [x for x in class_params_spec if x in self.transvars]

        betas_length = 0
        if 'class_intercept_alts' in self.intercept_opts and '_inter' in class_params_spec:
        # {
            # separate logic for intercept
            # QUERY. UNUSED CODE: class_isvars = [isvar for isvar in self.isvars if isvar != '_inter']
            betas_length += len(self.intercept_opts['class_intercept_alts'][class_num])
        # }
        else:  # {
            class_isvars = [x for x in class_params_spec if x in self.isvars]
            betas_length += (len(self.alts) - 1) * (len(class_isvars))
        # }

        betas_length += len(class_asvars)
        betas_length += len(class_randvars)

        # copied from choice model logic for Kchol
        betas_length = self.get_kchol(class_params_spec)
        betas_length += len(class_transvars) * 2
        betas_length += sum(self.rvtransidx)  # random trans vars
        betas_length += self.Kbw
        return betas_length
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Make a shortened dataframe                       '''
    ''' Average over alts used in latent class estimation         '''
    ''' ---------------------------------------------------------- '''
    def make_short_df(self, X):
    # {
        short_df = np.mean(np.mean(X, axis=2), axis=1)  # 2... over alts

        # Remove intercept columns
        if self.fit_intercept:  # {
            short_df = short_df[:, (self.J - 2):]
            short_df[:, 0] = 1
        # }

        # _________________________________________________
        if dev.using_gpu:
            short_df = dev.convert_array_cpu(short_df)
        # _________________________________________________

        self.short_df = short_df
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def set_bw(self, specs):
    # {
        specs = self.global_varnames[specs]
        self.varnames = specs
        randvars_specs = [param for param in specs if param in self.randvars]
        Kr = len(randvars_specs)
        self.Kbw, self.Kr = Kr, Kr      # Set self.Kbw and self.Kr as Kr
        self.Kftrans , self.Krtrans = sum(self.fxtransidx), sum(self.rvtransidx)
        self.rvdist = [dist for ii, dist in enumerate(self.global_rvdist) if self.randvars[ii] in randvars_specs]

        # Set up length of betas required to estimate correlation and/or
        # random variable standard deviations, useful for cholesky matrix
        if (self.correlated_vars):
        # {
            if (isinstance(self.correlated_vars, list)):  # {
                corvars_in_spec = [corvar for corvar in self.correlated_vars if corvar in randvars_specs]
                self.correlationLength = len(corvars_in_spec)
                self.Kbw = Kr - self.correlationLength
            # }
            else:  # {
                self.correlationLength, self.Kbw = Kr, 0
            # }
        # }
    # }

    def update_betas(self, betas):
        ''' FIx to make sure the betas hold for latent class'''
        iterations = [self.Kf, self.Kr, self.Kchol, self.Kbw, self.Kftrans,
                      self.Kftrans, self.Krtrans, self.Krtrans, self.Krtrans]
        if sum(iterations) != len(betas):
           
            missing_amount = sum(iterations) - len(betas)
            betas = np.append(betas, [0.91] * missing_amount)
        return betas

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def update(self, i, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs):
    # {
        #TODO FIX ME 4/11
        self.Kf, self.Kr = sum(class_fxidxs[i]), sum(class_rvidxs[i])
        self.fxidx, self.fxtransidx = class_fxidxs[i], class_fxtransidxs[i]
        self.rvidx, self.rvtransidx= class_rvidxs[i], class_rvtransidxs[i]

        # todo this need to be back respective to the intercept
        #FIXME
        #if self.intercept_classes[i] is False:
         #   class_idxs_sub = np.array([idx - (len(self.alts) - 2) for idx in class_idxs[i]])
        #else:
        class_idxs_sub = class_idxs[i]


        self.set_bw(class_idxs_sub)  # sets sd. and corr length
        self.Kchol = self.get_kchol(self.global_varnames[class_idxs[i]])


        rand_idx = [ii for ii, param in enumerate(self.randvars) if param in self.global_varnames[class_idxs_sub]]
        randtrans_idx = [ii for ii, param in enumerate(self.randtransvars) if param in self.global_varnames[class_idxs_sub]]
        return rand_idx, randtrans_idx
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def setup_em(self, X, y, class_thetas, class_betas):
    # {
        self.make_short_df(X)
        self.global_rvdist = self.rvdist
        self.global_varnames = self.varnames
        class_idxs, class_fxidxs, class_fxtransidxs, class_rvidx, class_rvtransidxs = self.setup_class()

        if '_inter' in self.global_varnames:
        # {
            for i in range(self.J - 2): #FIXME 5/11/24 adding _inter
                self.global_varnames = np.concatenate((np.array([f'_inter.{i}'], dtype='<U64'), self.global_varnames))
        # }

        if X.ndim != 4:  # {
            X = X.reshape(self.N, self.P, self.J, -1)
            y = y.reshape(self.N, self.P, self.J, -1)
        # }

        self.trans_pos = [ii for ii, var in enumerate(self.varnames) if var in self.transvars]  # used for get_class_X_idx

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # CLASS_THETAS
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if self.membership_as_probability:
            class_thetas = np.array([1 / (self.num_classes) for i in range(0, self.num_classes - 1)])

        if class_thetas is None and self.init_class_thetas is not None:
            class_thetas = self.init_class_thetas

        if class_thetas is None:
        # {
            len_class_thetas = [len(self.get_member_X_idx(i)) for i in range(0, self.num_classes - 1)]
            for ii, len_class_thetas_ii in enumerate(len_class_thetas): # {
                if '_inter' in self.member_params_spec[ii]:
                    len_class_thetas[ii] = len_class_thetas[ii] + 1
            # }
            class_thetas = np.concatenate([
                np.zeros(len_class_thetas[i])
                for i in range(0, self.num_classes - 1)], axis=0)
        # }

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # CLASS_BETAS
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if class_betas is None:
        # {
            class_betas = self.init_class_betas
            if class_betas is None:
                class_betas = [-0.1 * np.random.rand(self.get_betas_length(i)) for i in range(self.num_classes)]
        # }

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        return X, y, class_thetas, class_betas, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidx, class_rvtransidxs
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def setup_class(self):
    # {
        self.global_fxidx, self.global_fxtransidx= self.fxidx, self.fxtransidx
        self.global_rvidx,  self.global_rvtransidx = self.rvidx, self.rvtransidx
        class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs = [], [], [], [], []

        for class_num in range(self.num_classes):
        # {
        
            X_class_idx = self.get_class_X_idx(class_num) #FIXME 5 /11/24 change to one if broken
            class_idxs.append(X_class_idx)

            # deal w/ fix indices
            class_fx_idx = [fxidx for ii, fxidx in enumerate(self.fxidx) if ii in X_class_idx]
            class_fxtransidx = [fxtransidx for ii, fxtransidx in enumerate(self.fxtransidx) if ii in X_class_idx]

            # class_fxtransidx = np.repeat(False, len(X_class_idx))
            class_fxidxs.append(class_fx_idx)
            class_fxtransidxs.append(class_fxtransidx)

            # deal w/ random indices
            class_rv_idx = [rvidx for ii, rvidx in enumerate(self.rvidx) if ii in X_class_idx]
            class_rvtransidx = [rvtransidx for ii, rvtransidx in enumerate(self.rvtransidx) if ii in X_class_idx]

            # class_rvtransidx = np.repeat(False, len(X_class_idx))
            class_rvidxs.append(class_rv_idx)
            class_rvtransidxs.append(class_rvtransidx)
        # }
        return class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs
    # }

    def fixed_expectation_algorithm(self, tmp_fn, tmp_betas, args, class_thetas = None, class_betas = None, validation=False, **kwargs):
    # {

        X, y, panel_info, draws, drawstrans, weights, avail, batch_size = args
        if self.fixed_thetas is None:
            if self.fixed_solution is not None:
                self.fixed_thetas = self.fixed_solution['model'].class_x

        if self.fixed_thetas is not None:
            class_thetas = self.fixed_thetas

        X, y, class_thetas, class_betas, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs \
            = self.setup_em(X, y, class_thetas, class_betas)

       
        for s in range(0, self.num_classes):
            # {
            rand_idx, randtrans_idx = self.update(s, class_idxs, class_fxidxs,
                                                    class_fxtransidxs, class_rvidxs, class_rvtransidxs)
            updated_betas = self.update_betas(class_betas[s])
            class_betas[s] = updated_betas
       # np.random.uniform(0.23, 0.27, len(betas))

        #class_betas_sd = [np.repeat(0.25, len(betas)) for betas in class_betas]


        class_betas_sd = [np.random.uniform(0.23, 0.27, len(betas)) for betas in class_betas]
        if self.fixed_solution is not None:
            class_thetas_sd = self.fixed_solution['model'].class_x_stderr
        else:
            class_thetas_sd = np.repeat(0.01, class_thetas.size)
        log_lik_old, log_lik_new = -1E10, -1E10
        iter, max_iter = 0, 100
        terminate = False

        while not terminate and iter < max_iter:
            # {

            self.ind_pred_prob_classes = []
            self.choice_pred_prob_classes = []
            args = (0, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs)
            rand_idx, randtrans_idx = self.update(*args)
            args = (class_betas[0], X[:, :, :, class_idxs[0]], y, panel_info,
                    draws[:, rand_idx, :], drawstrans[:, randtrans_idx, :], avail)
            
            p = self.compute_probabilities_latent(*args)
        
            H = self.posterior_est_latent_class_probability(class_thetas)

            for i in range(1, self.num_classes):
                # {
                
                args = (i, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs)
                rand_idx, randtrans_idx = self.update(*args)
                args = (class_betas[i], X[:, :, :, class_idxs[i]], y, panel_info, draws[:, rand_idx, :],
                        drawstrans[:, randtrans_idx, :], avail)
                new_p = self.compute_probabilities_latent(*args)
                p = np.vstack((p, new_p))
            # }

            # ______________________________________
            if dev.using_gpu:
                p, H = dev.to_cpu(p), dev.to_cpu(H)
            # _____________________________________

            weights = np.multiply(p, H)
            weights[weights == 0] = min_comp_val
            log_lik = np.log(np.sum(weights, axis=0))  # sum over classes
            log_lik_new = np.sum(log_lik)
            weights_individual = weights
            tiled = np.tile(np.sum(weights_individual, axis=0), (self.num_classes, 1))
            weights_individual = np.divide(weights_individual, tiled)  # Compute weights_individual / tiled

            # NOTE: REMOVED CODE. MAY 2024. SEEMS REDUNDANT!
            # tiled = np.tile(np.sum(weights, axis=0), (self.num_classes, 1))
            # weights = np.divide(weights, tiled)

            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # SOLVE OPTIMISATION PROBLEM
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            converged = False


            self.pred_prob_all = np.array([])
            global_transvars = self.transvars.copy()

            self.panel_info = getattr(self, 'panel_info', None)  # i.e., Lookup or set as None

            for s in range(0, self.num_classes):
                # {
                rand_idx, randtrans_idx = self.update(s, class_idxs, class_fxidxs,
                                                      class_fxtransidxs, class_rvidxs, class_rvtransidxs)
                updated_betas = self.update_betas(class_betas[s])
                class_betas[s] = updated_betas
                #updates betas is longer than betas, how to copy updated betas into class_betas[s] so that 
                
                jac = True if self.return_grad else False  # QUERY: WHY IS THIS REQUIRED? WHY NOT USE self.jac
                self.total_fun_eval = 0

                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                # SOLVE OPTIMISATION PROBLEM
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                converged = False
                '''Dont think i need this'''
                """
                if self.intercept_classes[s]:
                    X_new = np.insert(X, 0, 1, axis=3)
                    args = (X_new[:, :, :,class_idxs[s]], y, self.panel_info, draws[:, rand_idx, :],
                        drawstrans[:, randtrans_idx, :], weights_individual[s, :], avail, batch_size)
                else:
                """
                args = (X[:, :, :, class_idxs[s]], y, self.panel_info, draws[:, rand_idx, :],
                        drawstrans[:, randtrans_idx, :], weights_individual[s, :], avail, batch_size)

                opt_res = minimize(self.get_loglik_gradient, class_betas[s], jac=jac, args=args, method="BFGS",
                                   tol=self.ftol, options={'gtol': self.gtol})
                """
                if self.intercept_classes[s]:
                    p = self.compute_probabilities(opt_res['x'], X_new[:, :, :, class_idxs[s]], panel_info,
                                                   draws[:, rand_idx, :], drawstrans[:, randtrans_idx, :], avail,
                                                   self.var_list, self.chol_mat)

                # save predicted and observed probabilities to display in summary
                else:
                """
                p = self.compute_probabilities(opt_res['x'], X[:, :, :, class_idxs[s]], panel_info,
                                               draws[:, rand_idx, :], drawstrans[:, randtrans_idx, :], avail,
                                               self.var_list, self.chol_mat)

                self.choice_pred_prob = np.mean(p, axis=3)
                self.ind_pred_prob = np.mean(self.choice_pred_prob, axis=1)
                self.pred_prob = np.mean(self.ind_pred_prob, axis=0)
                self.prob_full = p
                self.transvars = global_transvars
                self.pred_prob_all = np.append(self.pred_prob_all, self.pred_prob)
                self.ind_pred_prob_classes.append(self.ind_pred_prob)
                self.choice_pred_prob_classes.append(self.choice_pred_prob)

                if opt_res['success']:
                    # {
                    converged = True
                    class_betas[s] = opt_res['x']
                    prev_class_betas_sd = class_betas_sd

                    # Array tmp_calc contains the square roots of the absolute values
                    # of the diagonal elements of the inverse Hessian matrix.
                    tmp_calc = np.sqrt(np.abs(np.diag(opt_res['hess_inv'])))

                    # Compute the element-wise minimum between 0.25 * tmp_calc and
                    # prev_class_betas_sd[s], without the need for an explicit loop.
                    class_betas_sd[s] = np.minimum(0.25 * tmp_calc, prev_class_betas_sd[s])
                # }
            # }

            self.varnames = self.global_varnames

            terminate = log_lik_new - log_lik_old < self.ftol_lccmm

            # DEBUGGING:
            # print('class betas: ', class_betas)
            # print('class_thetas: ', class_thetas)
            # print(f'Loglik: {log_lik_new:.4f}')

            log_lik_old = log_lik_new
            iter += 1
            #class_thetas = class_thetas.reshape((self.num_classes - 1, -1))
        # }

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # This code concatenates arrays stored in the class_betas list into a single NumPy array
        x = np.concatenate(class_betas)
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        stderr = np.concatenate(class_betas_sd)
        optimisation_result = {'x': x, 'success': converged, 'fun': -log_lik_new, 'nit': iter,
                               'stderr': stderr, 'is_latent_class': True, 'class_x': class_thetas.flatten(),
                               'class_x_stderr': class_thetas_sd, 'hess_inv': opt_res['hess_inv']}

        self.fxidx, self.fxtransidx = self.global_fxidx, self.global_fxtransidx
        self.rvidx, self.rvtransidx = self.global_rvidx, self.global_rvtransidx
        self.varnames = self.global_varnames

        p_class = np.mean(H, axis=1)

        # --------------------------------------------------------
        if dev.using_gpu:
            self.pred_prob_all = dev.to_cpu(self.pred_prob_all)
        # ------------------------------------------------------

        self.pred_prob = np.zeros(self.J)
        for i in range(self.num_classes):  # {
            fr = i * self.J
            to = fr + self.J
            self.pred_prob += p_class[i] * self.pred_prob_all[fr:to]
        # }
        return optimisation_result
    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''


    def expectation_maximisation_algorithm(self, tmp_fn, tmp_betas, args,
                                           class_betas=None, class_thetas=None, validation=False, **kwargs):
        X, y, panel_info, draws, drawstrans, weights, avail, batch_size = args
        X, y, class_thetas, class_betas, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs \
            = self.setup_em(X, y, class_thetas, class_betas)
        

        for s in range(0, self.num_classes):
            # {
            rand_idx, randtrans_idx = self.update(s, class_idxs, class_fxidxs,
                                                    class_fxtransidxs, class_rvidxs, class_rvtransidxs)
            updated_betas = self.update_betas(class_betas[s])
            class_betas[s] = updated_betas


        class_betas_sd = [np.repeat(0.99, len(betas)) for betas in class_betas]
        class_thetas_sd = np.repeat(0.01, class_thetas.size)
        log_lik_old, log_lik_new = -1E10, -1E10
        iter, max_iter = 0, 2000
        terminate = False

        while not terminate and iter < max_iter:
            prev_converged= False
            # {
            #print("Iteration = ", iter, "(", max_iter, ")")
            self.ind_pred_prob_classes = []
            self.choice_pred_prob_classes = []
            args = (0, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs)
            rand_idx, randtrans_idx = self.update(*args)
            args = (class_betas[0], X[:, :, :, class_idxs[0]], y, panel_info,
                    draws[:, rand_idx, :], drawstrans[:, randtrans_idx, :], avail)

            # DEBUG: st = time()
            p = self.compute_probabilities_latent(*args)
            # DEBUG: end = time()
            # DEBUG: print(f"ComputeProbLatent: {end - st:.2f} seconds")

            H = self.posterior_est_latent_class_probability(class_thetas)

            for i in range(1, self.num_classes):
                # {
                args = (i, class_idxs, class_fxidxs, class_fxtransidxs, class_rvidxs, class_rvtransidxs)
                rand_idx, randtrans_idx = self.update(*args)
                args = (class_betas[i], X[:, :, :, class_idxs[i]], y, panel_info, draws[:, rand_idx, :],
                        drawstrans[:, randtrans_idx, :], avail)
                new_p = self.compute_probabilities_latent(*args)
                p = np.vstack((p, new_p))
            # }

            # ______________________________________
            if dev.using_gpu:
                p, H = dev.to_cpu(p), dev.to_cpu(H)
            # _____________________________________

            weights = np.multiply(p, H)
            weights[weights == 0] = min_comp_val
            log_lik = np.log(np.sum(weights, axis=0))  # sum over classes
            log_lik_new = np.sum(log_lik)
            weights_individual = weights
            tiled = np.tile(np.sum(weights_individual, axis=0), (self.num_classes, 1))
            weights_individual = np.divide(weights_individual, tiled)  # Compute weights_individual / tiled

            # NOTE: REMOVED CODE. MAY 2024. SEEMS REDUNDANT!
            # tiled = np.tile(np.sum(weights, axis=0), (self.num_classes, 1))
            # weights = np.divide(weights, tiled)

            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # SOLVE OPTIMISATION PROBLEM
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            converged = True
            opt_res = minimize(self.class_member_func, class_thetas.flatten(), jac=True,
                               args=(weights_individual, X), method='BFGS', tol=self.ftol,
                               options={'gtol': self.gtol_membership_func})
            #if opt_res['success'] or opt_res['status']== 2:  # {
            if opt_res['success']:
                #converged = True
                class_thetas = opt_res['x']
                prev_tmp_thetas_sd = class_thetas_sd
                tmp_thetas_sd = np.sqrt(np.abs(np.diag(opt_res['hess_inv'])))

                for ii, tmp_theta_sd in enumerate(tmp_thetas_sd):
                    # {
                    if prev_tmp_thetas_sd[ii] < 0.25 * tmp_theta_sd and prev_tmp_thetas_sd[ii] != 0.01 \
                            or np.isclose(tmp_thetas_sd[ii], 1.0):
                        tmp_thetas_sd[ii] = prev_tmp_thetas_sd[ii]
                # }
                class_thetas_sd = tmp_thetas_sd
            # }
            else: 
                converged = False

            self.pred_prob_all = np.array([])
            global_transvars = self.transvars.copy()

            self.panel_info = getattr(self, 'panel_info', None)  # i.e., Lookup or set as None

            for s in range(0, self.num_classes):
                # {
                rand_idx, randtrans_idx = self.update(s, class_idxs, class_fxidxs,
                                                      class_fxtransidxs, class_rvidxs, class_rvtransidxs)
                jac = True if self.return_grad else False  # QUERY: WHY IS THIS REQUIRED? WHY NOT USE self.jac
                self.total_fun_eval = 0

                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                # SOLVE OPTIMISATION PROBLEM
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                #converged = False
                '''OLD CODE:
                if self.intercept_classes[s]:
                    X_new = np.insert(X, 0, 1, axis=3)
                    Xslice = X_new[:, :, :,class_idxs[s]]
                else:
                    Xslice = X[:, :, :, class_idxs[s]]
                '''
                # NEW CODE
                class_index = class_idxs[s]  # Extract the relevant class indices

                # Check if intercept is needed and create the sliced array accordingly
                Xslice = X[:, :, :, class_index]
                #TODO flog this off for now. trying to get rid of the interepts
                '''
                Xslice = np.insert(X[:, :, :, class_index], 0, 1, axis=3) if self.intercept_classes[s] \
                    else X[:, :, :, class_index]
                    '''
                # END NEW CODE
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                args = (Xslice, y, self.panel_info, draws[:, rand_idx, :],
                        drawstrans[:, randtrans_idx, :], weights_individual[s, :], avail, batch_size)

                # DEBUG: st = time()
                opt_res = minimize(self.get_loglik_gradient, class_betas[s], jac=jac, args=args, method="BFGS",
                                   tol=self.ftol, options={'gtol': self.gtol})
                # DEBUG: end = time()
                # DEBUG: print(f"Minimize[{s}]: {end - st:.2f} seconds")

                # save predicted and observed probabilities to display in summary
                # DEBUG: st = time()

                ''' OLD CODE:
                if self.intercept_classes[s]:
                    Xslice = X_new[:, :, :, class_idxs[s]]
                else:
                    Xslice = X[:, :, :, class_idxs[s]]
                '''
                p = self.compute_probabilities(opt_res['x'], Xslice, panel_info,
                                               draws[:, rand_idx, :], drawstrans[:, randtrans_idx, :], avail, self.var_list,
                                               self.chol_mat)
                # DEBUG: end = time()
                # DEBUG:print(f"ComputeProb{s}: {end-st:.2f} seconds")

                self.choice_pred_prob = np.mean(p, axis=3)
                self.ind_pred_prob = np.mean(self.choice_pred_prob, axis=1)
                self.pred_prob = np.mean(self.ind_pred_prob, axis=0)
                self.prob_full = p
                self.transvars = global_transvars
                self.pred_prob_all = np.append(self.pred_prob_all, self.pred_prob)
                self.ind_pred_prob_classes.append(self.ind_pred_prob)
                self.choice_pred_prob_classes.append(self.choice_pred_prob)

                #if opt_res['success'] or opt_res['status']== 2:
                if opt_res['success'] or not prev_converged:
                    prev_class_betas_sd = opt_res['success']
                    # {
                    #converged = True
                    class_betas[s] = opt_res['x']
                    prev_class_betas_sd = class_betas_sd

                    # Array tmp_calc contains the square roots of the absolute values
                    # of the diagonal elements of the inverse Hessian matrix.
                    tmp_calc = np.sqrt(np.abs(np.diag(opt_res['hess_inv'])))

                    # Compute the element-wise minimum between 0.25 * tmp_calc and
                    # prev_class_betas_sd[s], without the need for an explicit loop.
                    class_betas_sd[s] = np.minimum(0.25 * tmp_calc, prev_class_betas_sd[s])
                else:
                    converged = False
            # }

            self.varnames = self.global_varnames
            terminate = np.abs(log_lik_new - log_lik_old) < self.ftol_lccmm
            if self.verbose > 1:
                print(f'Loglik: {log_lik_new:.4f}')
            # DEBUGGING:
            # print('class betas: ', class_betas)
            # print('class_thetas: ', class_thetas)
            # print(f'Loglik: {log_lik_new:.4f}')

            log_lik_old = log_lik_new
            iter += 1
            #FIX ME this falls over because it assumes we have same class sizes
            # TODO turning off for now, see if this holds up
            #class_thetas = class_thetas.reshape((self.num_classes - 1, -1))
        # }

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # This code concatenates arrays stored in the class_betas list into a single NumPy array
        x = np.concatenate(class_betas)
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        stderr = np.concatenate(class_betas_sd)
        optimisation_result = {'x': x, 'success': converged, 'fun': -log_lik_new, 'nit': iter,
                               'stderr': stderr, 'is_latent_class': True, 'class_x': class_thetas.flatten(),
                               'class_x_stderr': class_thetas_sd, 'hess_inv': opt_res['hess_inv']}

        self.fxidx, self.fxtransidx = self.global_fxidx, self.global_fxtransidx
        self.rvidx, self.rvtransidx = self.global_rvidx, self.global_rvtransidx
        self.varnames = self.global_varnames

        p_class = np.mean(H, axis=1)

        # --------------------------------------------------------
        if dev.using_gpu:
            self.pred_prob_all = dev.to_cpu(self.pred_prob_all)
        # ------------------------------------------------------

        self.pred_prob = np.zeros(self.J)
        for i in range(self.num_classes):
            # {
            fr = i * self.J
            to = fr + self.J
            self.pred_prob += p_class[i] * self.pred_prob_all[fr:to]
        # }
        return optimisation_result
        # }


    ''' ---------------------------------------------------------- '''
    ''' Function: Computes the log-likelihood on the validation set'''
    ''' using the betas fitted using the training set              '''
    ''' ---------------------------------------------------------- '''
    def validation_loglik(self, validation_X, validation_Y, panel_info=None, avail=None,
                          weights=None, panels=None,
                          betas=None, ids=None, batch_size=None, alts=None): # The inputs on this line are unused?
    # {
        N = len(np.unique(panels)) if panels is not None else self.N
        validation_X, Xnames = self.setup_design_matrix(validation_X)

        if len(np.unique(panels)) != (N / self.J):
        # {
            X, y, avail, panel_info = self.balance_panels(validation_X, validation_Y, avail, panels)
            validation_X = X.reshape((N, self.P, self.J, -1))
            validation_Y = y.reshape((N, self.P, self.J, -1))
        # }
        else:  # {
            validation_X = validation_X.reshape(N, self.P, self.J, -1)
            validation_Y = validation_Y.reshape(N, self.P, -1)
        # }

        batch_size = self.n_draws
        self.N = N  # store for use in EM alg
        # self.ids = ids

        draws, drawstrans = self.generate_draws(N, self.n_draws)  # (N,Kr,R)

        class_betas = []

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        counter = 0
        for _ in self.class_params_spec:  # {
            idx = counter + self.get_betas_length(0)
            class_betas.append(self.coeff_est[counter:idx])
            counter = idx
        # }

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        counter = 0
        for param_spec in self.member_params_spec:  # {
            idx = counter + len(param_spec)
            class_betas.append(self.coeff_est[counter:idx])
            counter = idx
        # }
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        tmp_fn = None
        tmp_betas = class_betas
        args = (validation_X, validation_Y, panel_info, draws, drawstrans, weights, avail, batch_size)
        res = self.expectation_maximisation_algorithm(tmp_fn, tmp_betas, args, validation=True)
        return res
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function: Override bfgs function                           '''
    ''' ---------------------------------------------------------- '''
    #todo this doesn't actually do anything because it doesnt overide anything, unlike latent class model
    def bfgs_optimization(self, betas, X, y, weights, avail, maxiter):  # {
        if self.optimise_class == True  and self.optimise_membership == False and self.fixed_solution is not None:
        # {
            thetas = self.fixed_solution['model'].class_x
            self.fixed_thetas = thetas
            # ERROR HERE? INPUT LIST IS ODD?
            result = self.fixed_expectation_algorithm(X, y, betas, class_thetas = thetas, validation=self.validation)
        # }
        else:
            result = self.expectation_maximisation_algorithm(X, y, avail, validation=self.validation)

        return result
    # }
# }
